//	CrystalFlightContentView.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI


let theMaxSpeedSliderWidth: CGFloat = 160.0

enum PanelType {
	case noPanel
	case crystalPanel
	case optionsPanel
	case exportPanel
	case helpPanel
	case contactPanel
	case translatorsPanel
	case visionOSPanel
}

//	The systemBackground color could change on the fly,
//	if the player switches between light mode and dark mode.
#if os(iOS)
var gSystemBackgroundColor: Color {
	Color(UIColor.systemBackground)
}
#endif
#if os(macOS)
var gSystemBackgroundColor: Color {
	Color(NSColor.textBackgroundColor)
}
#endif


struct CrystalFlightContentView: View {

	@State var itsModelData = CrystalFlightModel()
	@State var itsRenderer = CrystalFlightRenderer()
	@State var itsActivePanel: PanelType = .noPanel
	
	//	Unlike SwiftUI Gestures, which keep all the information they need,
	//	UIKit's UIGestureRecognizers don't keep strong references to their targets.
	//	So let's keep a reference to our CrystalFlightGestureHandler here.
	//	makePlatformDependentGestureRecognizers() will set the correct
	//	values for its instance variables.
	//
	//	It shouldn't create a strong reference cycle, but be aware
	//	that the itsCrystalFlightGestureHandler will hold a strong reference
	//	to the document.
	//
	//	If Crystal Flight someday migrates to SwiftUI gestures (when they
	//	support 2-finger drags), this @State variable can be removed.
	//
	@State var itsCrystalFlightGestureHandler = CrystalFlightGestureHandler()

	//	To make App Store screenshots, set gMakeAppStoreScreenshots = true.
	@State var itsScreenshotIndex: Int = 0	//	used only when making screenshots

	@State var itsCursorIsOverCrystalOnMacOS: Bool = false	//	ignored on iOS

	var body: some View {
					
		let theTranslatorThanks = [
			GeometryGamesTranslatorThanks(pre: "mit Dank an", name: "Stas Drisner", post: nil, lang: "de"),
			GeometryGamesTranslatorThanks(pre: "grazie a", name: "Daniela, Isabella Eva e Davide Bigatti", post: nil, lang: "it"),
			GeometryGamesTranslatorThanks(pre: "obrigado a", name: "Nuno Fernandes", post: nil, lang: "pt")
		]

		let theGestureRecognizers =
			itsCrystalFlightGestureHandler.makePlatformDependentGestureRecognizers(
				modelData: itsModelData,
				activePanel: $itsActivePanel,
				cursorIsOverCrystalOnMacOS: $itsCursorIsOverCrystalOnMacOS)

#if os(iOS)
		let theEdgesToExpandFromSafeArea: Edge.Set = .all
#endif
#if os(macOS)
		let theEdgesToExpandFromSafeArea: Edge.Set = .bottom
#endif

		return ZStack(alignment: .bottom) {
		
			//	Main animation view
			if let theRenderer = itsRenderer {
			
				GeometryGamesViewRep(
					modelData: itsModelData,
					renderer: theRenderer,
					gestureRecognizers: theGestureRecognizers
				)
				.ignoresSafeArea(edges: theEdgesToExpandFromSafeArea)
				.onHover() { cursorIsOverCrystal in
					itsCursorIsOverCrystalOnMacOS = cursorIsOverCrystal
				}
			}
			
			VStack() {

				//	At most one control panel
				Group() {
				
					//	Note:  Some of the panels get leading-aligned, others trailing-aligned.
					switch itsActivePanel {
					
					case .noPanel:
						Spacer()

					case .crystalPanel:
						CrystalChoiceView(modelData: itsModelData)
						.frame(maxWidth: .infinity, alignment: .leading)
						
					case .optionsPanel:
						OptionsView(modelData: itsModelData)
						.frame(maxWidth: .infinity, alignment: .leading)
					
					case .exportPanel:
						if let theRenderer = itsRenderer {
							GeometryGamesExportView(
								modelData: itsModelData,
								exportRenderer: theRenderer,
								caption: "Crystal Flight image")
							.frame(maxWidth: .infinity, alignment: .trailing)
						} else {
							Text("Renderer not available")
						}
						
					case .helpPanel:
						CrystalFlightHelpView()
						.frame(maxWidth: .infinity, alignment: .trailing)
						
					case .contactPanel:
						GeometryGamesContactView()
						.frame(maxWidth: .infinity, alignment: .trailing)
						
					case .translatorsPanel:
						GeometryGamesTranslatorsView(theTranslatorThanks)
						.frame(maxWidth: .infinity, alignment: .trailing)

					case .visionOSPanel:
						CrystalExplorerView()
						.frame(maxWidth: .infinity, alignment: .trailing)
						
					}
				}
				.padding(EdgeInsets(
							top: geometryGamesPanelMargin,
							leading: geometryGamesPanelMargin,
							bottom: 0.0,	//	OK to sit close to buttons
							trailing: geometryGamesPanelMargin))

				//	Various controls
				HStack(alignment: .bottom) {
				
					//	The buttons already have their own tappable padding,
					//	so the AdaptiveStack doesn't need to provide any extra spacing.
					AdaptiveStack(spacing: 0) {
						CrystalChoiceButton(activePanel: $itsActivePanel)
						OptionsButton(activePanel: $itsActivePanel)
					}
					
					Spacer()

					SpeedSlider(
						modelData: itsModelData,
						prePauseUserSpeed: $itsModelData.itsPrePauseUserSpeed)
				
					Spacer()

					//	The buttons already have their own tappable padding,
					//	so the AdaptiveStack doesn't need to provide any extra spacing.
					AdaptiveStack(spacing: 0) {
						ExportButton(activePanel: $itsActivePanel)
						HelpButton(activePanel: $itsActivePanel)
					}
				}
				.modifier(blueForegroundOnMacOS())
				.padding(EdgeInsets(
							top: 0.0,	//	OK to sit close to panel
							leading: geometryGamesControlInset,
							bottom: geometryGamesControlInset,
							trailing: geometryGamesControlInset))
			}
			.modifier(buttonStyleOnMacOS())
				
			if gMakeAppStoreScreenshots {
				//	The Rectangle will respond to taps
				//	only if it's at least a tiny bit opaque.
				Rectangle()
				.foregroundColor(Color(.displayP3,
									red: 0.0,
									green: 0.0,
									blue: 0.0,
									opacity: 0.000000000001))
				.ignoresSafeArea()
				.onTapGesture {
					itsScreenshotIndex = (itsScreenshotIndex + 1) % gNumAppStoreScreenshots
				}
			}
		}
//	Either make this iOS-only, or use KaleidoPaint's approach to hiding the status bar,
//	which seems to be to set "Status bar is initially hidden" in the info.plist file.
//		.statusBar(hidden: (gMakeWebSiteScreenshot || gMakeAppStoreScreenshots))
		.onAppear() {
			if gMakeWebSiteScreenshot {
				prepareForWebSiteScreenshot(
					modelData: itsModelData,
					activePanel: $itsActivePanel)
			}
			if gMakeAppStoreScreenshots {
				prepareForAppStoreScreenshot(
					itsScreenshotIndex,
					modelData: itsModelData,
					activePanel: $itsActivePanel)
			}
		}
		.onChange(of: itsScreenshotIndex) {
			prepareForAppStoreScreenshot(
				itsScreenshotIndex,
				modelData: itsModelData,
				activePanel: $itsActivePanel)
		}
	}
}


struct CrystalChoiceButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = activePanel == .crystalPanel ? .noPanel : .crystalPanel
		} label: {
			Image("Crystal A")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}


struct CrystalChoiceView: View {

	var modelData: CrystalFlightModel

	var body: some View {
	
		VStack(alignment: .trailing) {

			ForEach(Crystal.allCases, id: \.id) { theCrystal in

				Button(
					action: {
						modelData.itsCrystal = theCrystal
					},
					label: {
						HStack() {
						
							if theCrystal == modelData.itsCrystal {
								Text("✓")
							} else {
								Text("✓")
									.hidden()
							}
							
							VStack(alignment: .leading) {

								Text(LocalizedStringKey(theCrystal.familiarNameKey))
								.modifier(blueForegroundOnMacOS())
								.lineLimit(1)

								Text(verbatim: theCrystal.formula)
								.foregroundColor(.gray)
								.font(.footnote)
							}
							
							//	Caution:  This Spacer would make the CrystalChoiceView
							//	fill the full width of the display if we hadn't applied
							//	a .fixedSize modifier to its VStack.
							Spacer()
							
							Image("Crystal/" + theCrystal.formalNameKey)
						}
					}
				)
			}
		}
		.fixedSize()
		.padding(geometryGamesPanelPadding)
		.background(gSystemBackgroundColor)
		.cornerRadius(geometryGamesCornerRadius)
	}
}


struct OptionsButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = activePanel == .optionsPanel ? .noPanel : .optionsPanel
		} label: {
			Image(systemName: "switch.2")	//	or "slider.horizontal.3"
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}

struct OptionsView: View {

	@Bindable var modelData: CrystalFlightModel

	var body: some View {
	
		//	Set a fixed width for the fields showing
		//	the atomic radius factor and the horizon radius.
		//	Otherwise the whole view's width will jiggle
		//	as slightly wider or narrower numbers are displayed.
		let theValueFieldWidth: CGFloat = 64

		VStack() {
		
			//	Atomic radius factor
			VStack() {
				HStack() {
					Text("Atom Size")
					Spacer()
					Text("\(Int(modelData.itsAtomicRadiusFactor * 100.0))%")
						.frame(width: theValueFieldWidth, alignment: .trailing)
				}
				Slider(
					value: $modelData.itsAtomicRadiusFactor,
					in: gMinAtomicRadiusFactor ... gMaxAtomicRadiusFactor
				)
			}
			
			//	Caution:  Divider() will make the HStack fill all available width
			//	unless we apply .fixedSize() or .fixedSize(horizontal: vertical: )
			//	to the containing VStack().
			Divider()
			
			//	Horizon radius
			VStack() {
				HStack() {
					Text("Horizon Radius")
					Spacer()
					Text("\(modelData.itsHorizonRadius, specifier: "%.1lf") Å")
						.frame(width: theValueFieldWidth, alignment: .trailing)
				}
				Slider(
					value: $modelData.itsHorizonRadius,
					in: gMinHorizonRadius ... gMaxHorizonRadius)
			}
			
			Divider()	//	See Caution on Divider() above

			//	Bonds
			Toggle(isOn: $modelData.itsBondsAreVisible) {
				Text("Bonds")
			}

			//	Fog
			Toggle(isOn: $modelData.itsFogIsEnabled) {
				Text("Fog")
			}
		}
		.fixedSize()
		.padding(geometryGamesPanelPadding)
		.background(gSystemBackgroundColor)
		.cornerRadius(geometryGamesCornerRadius)
		.onChange(of: modelData.itsAtomicRadiusFactor) {
			modelData.changeCount += 1
		}
		.onChange(of: modelData.itsHorizonRadius) {
			//	Note:  Wait until the slider's onEditingChanged callback
			//	to call FinalizeHorizonRadius() to re-create the tiling.
			modelData.changeCount += 1
		}
		.onChange(of: modelData.itsBondsAreVisible) {
			modelData.changeCount += 1
		}
		.onChange(of: modelData.itsFogIsEnabled) {
			modelData.changeCount += 1
		}
	}
}

struct SpeedSlider: View {

	@Bindable var modelData: CrystalFlightModel
	@Binding var prePauseUserSpeed: Double?

	var body: some View {

		Slider(
			value: $modelData.itsUserSpeed,
			in: -gMaxUserSpeed ... +gMaxUserSpeed,
			onEditingChanged: {isEditing in
			
				//	If the user taps the main view to pause the forward motion,
				//	but then manually changes the speed before tapping
				//	the main view to unpause, we should set prePauseUserSpeed to nil
				//	to "cancel the paused status". Do this when the user lifts
				//	his/her finger from the Slider.
				
				if !isEditing {
					prePauseUserSpeed = nil
				}
			}
		)
		.frame(maxWidth: theMaxSpeedSliderWidth)
		.padding(geometryGamesControlPadding)
	}
}


struct ExportButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Button {
			activePanel = activePanel == .exportPanel ? .noPanel : .exportPanel
		} label: {
			Image(systemName: "square.and.arrow.up")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
	}
}


struct HelpButton: View {

	@Binding var activePanel: PanelType

	var body: some View {

		Menu() {

			Button() {
				activePanel = .helpPanel
			} label: {
				Label("How to fly", systemImage: "questionmark.circle")
			}
		
			Button() {
				activePanel = .contactPanel
			} label: {
				GeometryGamesHelpMenuContactItemLabel()
			}
			
			Button() {
				activePanel = .translatorsPanel
			} label: {
				GeometryGamesHelpMenuTranslatorsItemLabel()
			}

			Button() {
				activePanel = .visionOSPanel
			} label: {
				Label("Vision Pro", systemImage: "vision.pro")
			}
			
		} label: {

			Image(systemName: "questionmark.circle")
				.font(.title)
				.padding(geometryGamesControlPadding)
				.background(geometryGamesTappableClearColor)
		}
		.menuOrder(.automatic)
	}
}
