//	CrystalFlightGPUFunctions.metal
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

#include <metal_stdlib>
using namespace metal;
#include "CrystalFlightGPUDefinitions.h"	//	Include this *after* "using namespace metal"
											//		if you might need to use uint32_t in "…GPUDefinitions.h"

constant bool	gUseFog					[[ function_constant(0) ]];
constant bool	gMakeAppIcon			[[ function_constant(1) ]];
constant bool	gColorByLevelOfDetail	[[ function_constant(2) ]];


#pragma mark -
#pragma mark Transform and light meshes

kernel void CrystalFlightTransformationAndLightingFunction(
		  constant simd_float3x3			&aRotation				[[ buffer(BufferIndexRotation) ]],
		  constant float3					&aTranslation			[[ buffer(BufferIndexTranslation) ]],
	const device CrystalFlightUnlitVertex	*anUnlitVertexBuffer	[[ buffer(BufferIndexUnlitVertices) ]],
		  device CrystalFlightLitVertex		*aLitVertexBuffer		[[ buffer( BufferIndexLitVertices ) ]],
		  constant half3					&levelOfDetailColor		[[ buffer(BufferIndexLevelOfDetailColor),
																	function_constant(gColorByLevelOfDetail) ]],
		  ushort							aGridID					[[ thread_position_in_grid ]] )
{
	float3	tmpTranslatedPosition,
			tmpTranslatedAndRotatedPosition;
	half3	tmpRotatedNormal;
	half3	tmpRawColor;
	half	tmpDiffuseFactor,
			tmpSpecularFactor;
	half3	tmpShadedColor;

	//	The direction of maximal specular reflection is given by
	//	the pre-computed vector
	//
	//		normalize(theLightDirection + theViewerDirection)
	//
	half3	theLightDirection		= {-0.27h, +0.27h, -0.92h},	//	points towards the light source
		  //theViewerDirection		= { 0.00h,  0.00h, -1.00h},	//	points towards the observer
			theSpecularDirection	= {-0.14h, +0.14h, -0.98h};	//	direction of maximal specular reflection


	//	position
	
	tmpTranslatedPosition = anUnlitVertexBuffer[aGridID].pos + aTranslation;
	tmpTranslatedAndRotatedPosition = aRotation * tmpTranslatedPosition;
	
	aLitVertexBuffer[aGridID].pos = tmpTranslatedAndRotatedPosition;
	
	
	//	lighting

	if (gColorByLevelOfDetail)
	{
		tmpRawColor = levelOfDetailColor.rgb;
	}
	else
	{
		tmpRawColor = anUnlitVertexBuffer[aGridID].col;
	}

	tmpRotatedNormal = simd_half3x3(aRotation) * anUnlitVertexBuffer[aGridID].nor;

	if (gMakeAppIcon)
	{
		tmpDiffuseFactor  = 0.25h + 0.75h * max(0.0h, dot(theLightDirection, tmpRotatedNormal) );
		tmpSpecularFactor = 0.125h * pow(max(0.0h, dot(theSpecularDirection, tmpRotatedNormal) ), 16.0h);
	}
	else
	{
		tmpDiffuseFactor  = 0.50h + 0.50h * max(0.0h, dot(theLightDirection, tmpRotatedNormal) );
		tmpSpecularFactor = 0.25h * pow(max(0.0h, dot(theSpecularDirection, tmpRotatedNormal) ), 16.0h);
	}
	tmpShadedColor = tmpDiffuseFactor * tmpRawColor
				   + half3(tmpSpecularFactor, tmpSpecularFactor, tmpSpecularFactor);
	
	aLitVertexBuffer[aGridID].col = tmpShadedColor;
}


#pragma mark -
#pragma mark Create tiling and do frustum culling

kernel void CrystalFlightTileCullingFunction(
	constant CrystalFlightTilingParameters	&someParams	[[ buffer(BufferIndexTilingParameters)	]],
	device   CrystalFlightTile				*aTiling	[[ buffer(BufferIndexCulledTiling)		]],
			 ushort3						aGridID		[[ thread_position_in_grid				]] )
{
	float3	theGridIndices,
			theTranslationVector,
			theTranslatedBoundingSphereCenter;
	int		theLevelOfDetail;
	float	theDistance;
	uint	theTileIndex,
			theOutputIndex;
	
	theGridIndices = float3(someParams.itsGridLowerLimits + short3(aGridID));
	theTranslationVector = someParams.itsGridTranslations * theGridIndices;
	theTranslatedBoundingSphereCenter = someParams.itsBoundingSphereCenter + theTranslationVector;

	theLevelOfDetail = -1;

	//	Does the bounding sphere survive frustum culling?
	if (
		dot(someParams.itsOutwardUnitNormals[0], theTranslatedBoundingSphereCenter)
			< someParams.itsBoundingSphereRadius
	 &&
		dot(someParams.itsOutwardUnitNormals[1], theTranslatedBoundingSphereCenter)
			< someParams.itsBoundingSphereRadius
	 &&
		dot(someParams.itsOutwardUnitNormals[2], theTranslatedBoundingSphereCenter)
			< someParams.itsBoundingSphereRadius
	 &&
		dot(someParams.itsOutwardUnitNormals[3], theTranslatedBoundingSphereCenter)
			< someParams.itsBoundingSphereRadius
	)
	{
		//	The output buffer (aTiling) contains four contiguous arrays
		//	of CrystalFlightTiles, one array for each level of detail,
		//	indexed as follows:
		//
		//		 0  coarsest level of detail
		//		 1  ...
		//		 2  ...
		//		 3  finest level of detail
		//
		theDistance = length(theTranslatedBoundingSphereCenter);

		if (theDistance >= someParams.itsAdjustedHorizonRadius) {
			//	The tile is too far away, so leave
			//	theLevelOfDetail set to -1 to cull this tile.
		} else if (theDistance > someParams.itsDistanceCutoffs[0]) {
			theLevelOfDetail = 0;	//	coarsest level  (for distant copies of the unit cell)
		} else if (theDistance > someParams.itsDistanceCutoffs[1]) {
			theLevelOfDetail = 1;
		} else if (theDistance > someParams.itsDistanceCutoffs[2]) {
			theLevelOfDetail = 2;
		} else {
			theLevelOfDetail = 3;	//	 finest level   (for nearby copies of the unit cell)
		}
	}

	if (theLevelOfDetail >= 0) {
	
		theTileIndex = uint(aGridID[0])
							* uint(someParams.itsGridDimensions[1])
							* uint(someParams.itsGridDimensions[2])
					 + uint(aGridID[1])
							* uint(someParams.itsGridDimensions[2])
					 + uint(aGridID[2]);
		
		theOutputIndex = uint(theLevelOfDetail) * someParams.itsTileCount
					   + theTileIndex;

		aTiling[theOutputIndex] = CrystalFlightTile(
									half3(theTranslationVector),
									1.0);
	}
}


#pragma mark -
#pragma mark Compact the tiling arrays and prepare indirect argument buffers

kernel void CrystalFlightTilingArrayCompactionFunction(
	device   CrystalFlightTile	*aTiling			[[ buffer(BufferIndexCompactableTiling)	]],
	constant ushort				&aLoopCount			[[ buffer(BufferIndexLoopCount)			]],
	device   MTLDrawIndexedPrimitivesIndirectArguments
					*someIndirectArguments			[[ buffer(BufferIndexIndirectArguments)	]],
			 ushort2			aThreadgroupSize	[[ threads_per_threadgroup				]],
			 ushort2			aGridID				[[ thread_position_in_grid				]] )
{
	ushort						theLevelOfDetail,
								theNumTilesPerLoopPass,
								theNumTilesPerLevel;
	uint						theLevelOfDetailIndex;
	device CrystalFlightTile	*theLevelOfDetailSubtiling;
	ushort						theTileIndexInLoopPass,
								theLoopPass;

	//	The array b[] lives in threadgroup memory.
	//	We'll initialize it to contain a 1 for each tile
	//	that survived frustum culling, and a 0 for each tile
	//	that didn't.  Then we'll run the Hillis-Steele algorithm
	//	on it to get a running sum of the number of surviving tiles.
	//	Here's a simple example showing the log₂(8) = 3 steps
	//	of the Hillis-Steele algorithm acting on an 8-element array:
	//
	//		1	1	0	0	0	1	1	0
	//		1	2	1	0	0	1	2	1	(theOffset = 1)
	//		1	2	2	2	1	1	2	2	(theOffset = 2)
	//		1	2	2	2	2	3	4	4	(theOffset = 4)
	//
	threadgroup ushort	b[gCompactionThreadgroupMaxHeight];

	//	Keep track of how many non-culled CrystalFlightTiles
	//	we've found and written to their compacted positions in aTiling.
	threadgroup uint	theTileCount;

	//	Recall from the CrystalFlightTileCullingFunction above
	//	that the output buffer (aTiling) contains four contiguous arrays
	//	of CrystalFlightTiles, one array for each level of detail.
	//	aGridID[0] tells which of those arrays we're compacting here.
	
	theLevelOfDetail = aGridID[0];
	theNumTilesPerLoopPass = aThreadgroupSize[1];
	theNumTilesPerLevel = theNumTilesPerLoopPass * aLoopCount;
	theLevelOfDetailIndex = theLevelOfDetail * theNumTilesPerLevel;
	theLevelOfDetailSubtiling = aTiling + theLevelOfDetailIndex;
	
	if (aGridID[1] == 0)
		theTileCount = 0;
	threadgroup_barrier(mem_flags::mem_threadgroup);

	//	The threadgroup as a whole will process one level of detail.
	//	Each call to this kernel function will process a single tile at each level.
	theTileIndexInLoopPass = aGridID[1];
	
	for (theLoopPass = 0; theLoopPass < aLoopCount; theLoopPass++)
	{
		//	Read theTile and write a 1 into b[] if theTile
		//	survived culling or a 0 if it did not.
		CrystalFlightTile theTile
			= theLevelOfDetailSubtiling[theLoopPass * theNumTilesPerLoopPass + theTileIndexInLoopPass];
		b[theTileIndexInLoopPass] = (theTile[3] != 0.0 ? 1 : 0);
		threadgroup_barrier(mem_flags::mem_threadgroup);
		
		//	Use the Hillis-Steele algorithm to compute
		//	a running total of the numbers in the b[] array.
		for (ushort theOffset = 1; theOffset < theNumTilesPerLoopPass; theOffset *= 2)
		{
			ushort	theIncrement = (theTileIndexInLoopPass >= theOffset ?
										b[theTileIndexInLoopPass - theOffset] :
										0.0);
			threadgroup_barrier(mem_flags::mem_threadgroup);
		
			b[theTileIndexInLoopPass] += theIncrement;
			threadgroup_barrier(mem_flags::mem_threadgroup);
		}
		
		//	If theTile is one that survived culling,
		//	write it to its compacted position in aTiling.
		if (theTile[3] != 0.0)
			theLevelOfDetailSubtiling[theTileCount + (b[theTileIndexInLoopPass] - 1)] = theTile;
		threadgroup_barrier(mem_flags::mem_threadgroup);

		//	Update theTileCount.
		if (aGridID[1] == 0)
			theTileCount += b[theNumTilesPerLoopPass - 1];
		threadgroup_barrier(mem_flags::mem_threadgroup);
	}

	if (aGridID[1] == 0)
	{
		someIndirectArguments[0 + theLevelOfDetail].instanceCount = theTileCount;	//	for atoms
		someIndirectArguments[4 + theLevelOfDetail].instanceCount = theTileCount;	//	for bonds
	}
}


#pragma mark -
#pragma mark Render

struct VertexInput
{
	float3	pos [[ attribute(VertexAttributePosition)	]];
	half3	col [[ attribute(VertexAttributeColor)		]];	//	linear extended-range sRGB
};

struct VertexOutput
{
	float4	position		[[ position			]];
	float	clipDistance	[[ clip_distance	]] [1];
	half3	color			[[ user(color)		]];	//	linear extended-range sRGB
};
struct FragmentInput
{
	half3	color			[[ user(color)		]];	//	linear extended-range sRGB
};


vertex VertexOutput CrystalFlightVertexFunction(
	VertexInput							in					[[ stage_in						]],
	constant CrystalFlightUniformData	&uniforms			[[ buffer(BufferIndexUniforms)	]],
	const device CrystalFlightTile		*tilingGroup		[[ buffer(BufferIndexTiling)	]],
	uint								iid					[[ instance_id					]] )
{
	VertexOutput	out;
	float3			theTranslatedPosition;
	half			theRadius;
	half			theFogIntensity;
	half3			theFoggedColor;

	//	position
	
	//	tilingGroup[iid] holds the pre-computed product of the tiling matrix and the view matrix.
	theTranslatedPosition = in.pos + float3(tilingGroup[iid].xyz);

	out.position = uniforms.itsProjectionMatrix * float4(theTranslatedPosition, 1.0);
	
	//	radial clipping
	//
	//	Clip away any content that lies beyond the horizon radius,
	//	to provide a clean boundary when drawing with fog disabled
	//	(and also a tiny performance improvement when drawing with fog enabled).
	//
	theRadius			= length(half3(theTranslatedPosition));
	out.clipDistance[0]	= float( uniforms.itsHorizonRadius - theRadius );

	//	color

	if (gUseFog)
	{
		//	Let theFogIntensity be zero when theRadius is less than itsFogBegin,
		//	to keep foreground atoms maximally bright.
		//	Then let theFogIntensity increase linearly from itsFogBegin to the horizon.
		//	For efficiency, the CPU code has precomputed
		//
		//		                                   1
		//		itsFogRangeInverse = ------------------------------
		//		                     itsHorizonRadius - itsFogBegin
		//
		theFogIntensity = clamp((theRadius - uniforms.itsFogBegin) * uniforms.itsFogRangeInverse, 0.0h, 1.0h);
		theFoggedColor	= (1.0h - theFogIntensity) * in.col.rgb;
	}
	else	//	! gUseFog
	{
		theFoggedColor	= in.col.rgb;
	}

	out.color = theFoggedColor;

	return out;
}

fragment half4 CrystalFlightFragmentFunction(
	FragmentInput	in	[[ stage_in ]])
{
	return half4(in.color, 1.0h);
}
