//	CrystalFlightGestureHandler.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI


@MainActor
class CrystalFlightGestureHandler {

	//	makePlatformDependentGestureRecognizers()
	//	will set the correct values.
	var modelData: CrystalFlightModel = CrystalFlightModel()			//	dummy value
	var activePanel: Binding<PanelType> = .constant(.noPanel)			//	dummy value
	var cursorIsOverCrystalOnMacOS: Binding<Bool> = .constant(false)	//	dummy value

	func makePlatformDependentGestureRecognizers(
		modelData: CrystalFlightModel,
		activePanel: Binding<PanelType>,
		cursorIsOverCrystalOnMacOS: Binding<Bool>	//	ignored on iOS
	) -> [GeometryGamesPlatformDependentGestureRecognizer] {
	
		self.modelData = modelData
		self.activePanel = activePanel
		self.cursorIsOverCrystalOnMacOS = cursorIsOverCrystalOnMacOS
		
#if os(macOS)
		enableMacOSScrollEvents()
#endif

		var theGestureRecognizers: [GeometryGamesPlatformDependentGestureRecognizer] = []

		//	thePanGestureRecognizer interprets
		//
		//		1-finger drags as rotations about the x- and y-axes, and
		//		2-finger drags as translations.
		//
		let theOneFingerDragGestureRecognizer
			= GeometryGamesPlatformDependentPanGestureRecognizer(
				target: self,
				action: #selector(handleOneFingerDragGesture))
#if os(iOS)
		theOneFingerDragGestureRecognizer.minimumNumberOfTouches = 1
		theOneFingerDragGestureRecognizer.maximumNumberOfTouches = 1
#endif
		theOneFingerDragGestureRecognizer.delegate = modelData.itsGestureRecognizerDelegate
		theGestureRecognizers.append(theOneFingerDragGestureRecognizer)

		//	theRotationGestureRecognizer interprets
		//
		//		2-finger rotations as rotations about the z-axis.
		//
		let theRotationGestureRecognizer
			= GeometryGamesPlatformDependentRotationGestureRecognizer(
				target: self,
				action: #selector(handleRotationGesture))
		theRotationGestureRecognizer.delegate = modelData.itsGestureRecognizerDelegate
		theGestureRecognizers.append(theRotationGestureRecognizer)

		//	thePinchGestureRecognizer interprets
		//
		//		2-finger pinches as forward and backward motions.
		//
		let thePinchGestureRecognizer
			= GeometryGamesPlatformDependentPinchGestureRecognizer(
				target: self,
				action: #selector(handlePinchGesture))
		thePinchGestureRecognizer.delegate = modelData.itsGestureRecognizerDelegate
		theGestureRecognizers.append(thePinchGestureRecognizer)

#if os(iOS)
		let theTwoFingerDragGestureRecognizer = UIPanGestureRecognizer(
			target: self,
			action: #selector(handleTwoFingerDragGesture))
		theTwoFingerDragGestureRecognizer.minimumNumberOfTouches = 2
		theTwoFingerDragGestureRecognizer.maximumNumberOfTouches = 2
		theTwoFingerDragGestureRecognizer.delegate = modelData.itsGestureRecognizerDelegate
		theGestureRecognizers.append(theTwoFingerDragGestureRecognizer)
#endif

		//	theTapGestureRecognizer interprets
		//
		//		a single 1-finger tap
		//			as pausing or resuming the user's motion.
		//
		let theTapGestureRecognizer
			= GeometryGamesPlatformDependentTapGestureRecognizer(
				target: self,
				action: #selector(handleTapGesture))
#if os(iOS)
		theTapGestureRecognizer.numberOfTapsRequired = 1	//	how many taps
		theTapGestureRecognizer.numberOfTouchesRequired = 1	//	...with how many fingers
#endif
#if os(macOS)
		theTapGestureRecognizer.numberOfClicksRequired = 1
#endif
		theTapGestureRecognizer.delegate = modelData.itsGestureRecognizerDelegate
		theGestureRecognizers.append(theTapGestureRecognizer)

		return theGestureRecognizers
	}


// MARK: -
// MARK: macOS scroll events

#if os(macOS)
	func enableMacOSScrollEvents() {
	
		//	macOS treats a 2-finger drag on the trackpad not as a gesture,
		//	but rather as a scrollWheel event, presumably because back
		//	in the old days Macs had no trackpads but some mice had
		//	a little 1D scroll wheel between their left and right buttons,
		//	which later evolved to a 2D scroll wheel.
		
		NSEvent.addLocalMonitorForEvents(matching: [.scrollWheel], handler: { event in
		
			if event.type == .scrollWheel
			&& self.cursorIsOverCrystalOnMacOS.wrappedValue {
			
				let theUnscaledTranslationVector
					= SIMD3<Double>(	//	undefined units
						-event.scrollingDeltaX,
						 event.scrollingDeltaY,	//	flip vertically
						 0.0)

				let ångströmsPerScrollUnit = 0.005
				let theScaledTranslationVector = ångströmsPerScrollUnit * theUnscaledTranslationVector
					
				let theRotatedTranslationVector
					= self.modelData.itsUserOrientation.act(theScaledTranslationVector)

				self.modelData.itsUserPosition += theRotatedTranslationVector

				self.modelData.changeCount += 1

				if gGetScreenshotOrientations {
					self.modelData.printUserBodyPlacementToConsole()
				}
			}
			
			return event
		})
	}
#endif


// MARK: -
// MARK: 1- and 2-finger drag gestures

	@objc func handleOneFingerDragGesture(
		_ panGestureRecognizer: GeometryGamesPlatformDependentPanGestureRecognizer
	) {
	
		//	The compiler doesn't check the type
		//	of #selector(handleOneFingerDragGesture),
		//	so let's check it at run time.
		guard panGestureRecognizer.isKind(of:
			GeometryGamesPlatformDependentPanGestureRecognizer.self)
		else { return }
		
		handleDragGesture(panGestureRecognizer, purpose: .toRotate)
	}

#if os(iOS)
	@objc func handleTwoFingerDragGesture(
		_ panGestureRecognizer: UIPanGestureRecognizer
	) {

		//	The compiler doesn't check the type
		//	of #selector(handleTwoFingerDragGesture),
		//	so let's check it at run time.
		guard panGestureRecognizer.isKind(of:
			UIPanGestureRecognizer.self)
		else { return }
		
		handleDragGesture(panGestureRecognizer, purpose: .toTranslate)
	}
#endif	//	os(iOS)
#if os(macOS)
	//	On macOS we use scrollWheel to handle 2-finger drags.
#endif

	enum DragPurpose {
		case toRotate
		case toTranslate
	}

	func handleDragGesture(
		_ panGestureRecognizer: GeometryGamesPlatformDependentPanGestureRecognizer,
		purpose: DragPurpose
	) {

		guard
			let theGestureLocation
				= getGestureLocation(gestureRecognizer: panGestureRecognizer),
			let theGestureMotion
				= getPanGestureTranslation(panGestureRecognizer: panGestureRecognizer)
		else { return }

		switch panGestureRecognizer.state {

		case .possible:
			break
			
		case .began,	//	Respond to the initial pan when the gesture is first recognized
			.changed:	//	... as well as to subsequent panning

			//	We want incremental translations, not cumulative ones,
			//	so reset the panGestureRecognizer's translation to 0.0.
			panGestureRecognizer.setTranslation(CGPoint.zero, in: nil)
			
			//	The characteristic view size is the distance in the view,
			//	measured from the center of the view outwards,
			//	that subtends a 45° angle in the observer's field-of-view.
			let theCharacteristicSize = characteristicViewSize(
				frameWidth:  theGestureLocation.viewWidth,
				frameHeight: theGestureLocation.viewHeight)
			
			if theCharacteristicSize <= 0.0 {	//	should never occur
				assertionFailure("nonpositive characteristic size")
				return
			}

			//	Express theGestureLocation and theGestureMotion
			//	as multiples of the characteristic size.
			let theConversionFactor = 1.0 / theCharacteristicSize
			let x = theConversionFactor
				  * (theGestureLocation.x  -  0.5 * theGestureLocation.viewWidth )
			let y = theConversionFactor
				  * (theGestureLocation.y  -  0.5 * theGestureLocation.viewHeight)
			let Δx = theConversionFactor * theGestureMotion.Δx
			let Δy = theConversionFactor * theGestureMotion.Δy

			switch purpose {
			
			case .toRotate:
			
				//	In Crystal Flight, the characteristic size is,
				//	by definition, the distance on the display
				//	(measured from the center of the view outwards)
				//	that subtends a 45° angle in the simulated space.
				//	Thus we may imagine the user to be sitting
				//	one characteristic unit in front of the center
				//	of the display.
				
				let thePointA = SIMD3<Double>(x - Δx, y - Δy, 1.0)	//	in characteristic units
				let thePointB = SIMD3<Double>(  x,      y,    1.0)	//	in characteristic units
				
				//	We want to rotate the user in the sense opposite
				//	to that in which the user is "rotating the scenery",
				//	so define theRotation not from thePointA to thePointB,
				//	but the other way around.
				let theRotation = simd_quatd(from: thePointB, to: thePointA)
				
				modelData.itsUserOrientation = modelData.itsUserOrientation * theRotation
			
			case .toTranslate:
			
				//	Fudge alert:
				//	(Δx, Δy) give a well defined angle, but how that angle
				//	corresponds to a translation distance depends on how far way
				//	the objects that the user is focusing on are.
				//	In practice, multiplying by 4 seems to work well
				//	for the crystals in the Crystal Flight collection.
				let theTranslationFudgeFactor = 4.0	//	in ångströms per characteristic size unit
				
				//	Note:
				//		Use -Δx and -Δy to translate the scenery,
				//		or  +Δx and +Δy to translate the observer.
				//
				let theRawTranslationVector = SIMD3<Double>(
					-theTranslationFudgeFactor * Δx,
					-theTranslationFudgeFactor * Δy,
					0.0)
					
				let theRotatedTranslationVector
					= modelData.itsUserOrientation.act(theRawTranslationVector)

				modelData.itsUserPosition += theRotatedTranslationVector
			}

		case .ended:
			if gGetScreenshotOrientations {
				modelData.printUserBodyPlacementToConsole()
			}

		case .cancelled:
			break

		case .failed:
			break
			
		@unknown default:
			break
		}
		
		modelData.changeCount += 1
	}


// MARK: -
// MARK: Rotation gesture

	@objc func handleRotationGesture(
		_ rotationGestureRecognizer: GeometryGamesPlatformDependentRotationGestureRecognizer
	) {

		//	The compiler doesn't check the type
		//	of #selector(handleRotationGesture),
		//	so let's check it at run time.
		guard rotationGestureRecognizer.isKind(of: GeometryGamesPlatformDependentRotationGestureRecognizer.self)
		else { return }

		switch rotationGestureRecognizer.state {

		case .possible:
			break
			
		case .began,	//	Includes the initial small rotation at time of gesture recognition
			.changed:
			
			//	The SwiftUI RotateGesture() sometimes reports θ = NaN.
			//	I'm not sure whether the UIRotationGestureRecognizer
			//	and NSRotationGestureRecognizer also have this bug,
			//	but just to be safe let's check.
			let theRawAngle = rotationGestureRecognizer.rotation
			if theRawAngle.isNaN {
				return
			}

			//	We want incremental rotation angles, not cumulative ones,
			//	so reset the rotationGestureRecognizer's rotation angle to zero.
			rotationGestureRecognizer.rotation = 0.0

#if os(iOS)
			let θ = +Double(theRawAngle)
#endif
#if os(macOS)
			let θ = -Double(theRawAngle)
#endif

			let theRotation = simd_quatd(
								angle: θ,
								axis: SIMD3<Double>(0.0, 0.0, 1.0))

			modelData.itsUserOrientation = modelData.itsUserOrientation * theRotation

		case .ended:
			if gGetScreenshotOrientations {
				modelData.printUserBodyPlacementToConsole()
			}

		case .cancelled:
			break

		case .failed:
			break
			
		@unknown default:
			break
		}
		
		modelData.changeCount += 1
	}


// MARK: -
// MARK: Pinch gesture

	@objc func handlePinchGesture(
		_ pinchGestureRecognizer: GeometryGamesPlatformDependentPinchGestureRecognizer
	) {

		//	The compiler doesn't check the type
		//	of #selector(handlePinchGesture),
		//	so let's check it at run time.
		guard pinchGestureRecognizer.isKind(of:
			GeometryGamesPlatformDependentPinchGestureRecognizer.self)
		else { return }

		switch pinchGestureRecognizer.state {

		case .possible:
			break
			
		case .began,	//	Includes the initial small pinching at time of gesture recognition
			.changed:

			//	We want incremental pinches, not cumulative ones.
#if os(iOS)
			let theExpansionFactor = pinchGestureRecognizer.scale
			pinchGestureRecognizer.scale = 1.0
#endif
#if os(macOS)
			let theExpansionFactor = 1.0 + pinchGestureRecognizer.magnification
			pinchGestureRecognizer.magnification = 0.0
#endif

			//	Conceptually, there are various ways that we might map
			//	a pinch motion to a forward/backward motion.
			//	For example, we might translate by log(anExpansionFactor)
			//	or by (1 - anExpansionFactor)/anExpansionFactor .
			//	In practice, though, anExpansionFactor is
			//	an incremental expansion factor (representing the change
			//	since the last UIGestureRecognizerStateChanged event)
			//	and is therefore always close to 1.0.
			//	So no matter what conceptual model we might use,
			//	near 1.0 its effect will essentially be that of its linear approximation.
			//	So rather than sweating over the details of the conceptual model
			//	(which might require knowing the absolute spacing between the user's fingers)
			//	let's just skip straight to the linear approximation.
			//	The only choice we need to make is the PINCH_TO_TRANSLATION_FACTOR.
			//
			let thePinchToTranslationFactor = 8.0	//	ångströms per change in expansion factor
			let theRawTranslationVector = SIMD3<Double>(
				0.0,
				0.0,
				thePinchToTranslationFactor * (theExpansionFactor - 1.0))
				
			let theRotatedTranslationVector
				= modelData.itsUserOrientation.act(theRawTranslationVector)
			
			modelData.itsUserPosition += theRotatedTranslationVector

		case .ended:
			if gGetScreenshotOrientations {
				modelData.printUserBodyPlacementToConsole()
			}

		case .cancelled:
			break

		case .failed:
			break
			
		@unknown default:
			break
		}
		
		modelData.changeCount += 1
	}


// MARK: -
// MARK: Tap gesture

	@objc func handleTapGesture(
		_ tapGestureRecognizer: GeometryGamesPlatformDependentTapGestureRecognizer
	) {

		//	The compiler doesn't check the type
		//	of #selector(handleTapGesture),
		//	so let's check it at run time.
		guard tapGestureRecognizer.isKind(of:
			GeometryGamesPlatformDependentTapGestureRecognizer.self)
		else { return }

		switch tapGestureRecognizer.state {
		
		case .recognized:
		
			//	If a temporary panel is active, rather than interpreting
			//	the tap normally, interpret it as a request to dismiss
			//	the temporary panel.
			if activePanel.wrappedValue != .noPanel {
				activePanel.wrappedValue = .noPanel
				break
			}

			//	No temporary panel is active, so interpret the tap normally,
			//	as a request to pause or unpause the motion.
			if let thePrePauseUserSpeed = modelData.itsPrePauseUserSpeed {

				//	Unpause the forward motion
				
				//	Restore the speed as it was when the user last paused the motion.
				modelData.itsUserSpeed = thePrePauseUserSpeed
				modelData.itsPrePauseUserSpeed = nil
				
			} else {	//	itsPrePauseUserSpeed == nil
			
				//	Pause the forward motion
				
				//	Remember the current speed and pause the motion.
				modelData.itsPrePauseUserSpeed = modelData.itsUserSpeed
				modelData.itsUserSpeed = 0.0
			}

			if gGetScreenshotOrientations {
				modelData.printUserBodyPlacementToConsole()
			}

		default:
			//	We don't expect any other messages,
			//	but if any did arrive we'd ignore them.
			break
		}
		
		modelData.changeCount += 1
	}
}


// MARK: -
// MARK: Gesture recognizer delegate

class CrystalFlightGestureRecognizerDelegate: NSObject,
	GeometryGamesPlatformDependentGestureRecognizerDelegate {
	
	func gestureRecognizer(
		_ gestureRecognizer: GeometryGamesPlatformDependentGestureRecognizer,
		shouldRecognizeSimultaneouslyWith otherGestureRecognizer:
			GeometryGamesPlatformDependentGestureRecognizer
	) -> Bool {

		//	Let all gesture recognizers run simultaneously.
		return true
	}
}
