//	CrystalFlightMeshSet.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import Metal


let gNumLevelsOfDetail = 4

//	Using 16-bit integers to index mesh vertices works great
//	with the current implementation of Crystal Flight, and
//	makes good use of the GPU's power-efficient 16-bit arithmetic.
//	But if we ever needed to create meshes with 2¹⁶ = 65536
//	or more vertices (for example for extra-smooth spheres),
//	we'd need to index the mesh vertices with 32-bit integers instead.
typealias CrystalFlightMeshIndexType = UInt16	//	UInt16 or UInt32


struct Sphere {

	let center: SIMD3<Double>
	let radius: Double
	
	func encloses(_ someOtherSphere: Sphere) -> Bool {
	
		return (
			length(someOtherSphere.center - self.center)
		  + someOtherSphere.radius
		 <= self.radius )
	}
}


struct CrystalFlightUnlitMeshSet {

	let itsDevice: MTLDevice

	var itsPreparedCrystal: Crystal {
		didSet {
			if itsPreparedCrystal != oldValue {
				makeUnitCellAtomsMeshSet()
				makeUnitCellBondsMeshSet()
				computeBoundingSphere()
			}
		}
	}
	
	var itsPreparedAtomicRadiusFactor: Double {
		didSet {
			if itsPreparedAtomicRadiusFactor != oldValue {
				makeUnitCellAtomsMeshSet()
				computeBoundingSphere()
			}
		}
	}

	//	itsAtomsMeshSet (resp. itsBondsMeshSet) stores
	//	several meshes, each representing the atoms (resp. bonds)
	//	within a single copy of the unit cell,
	//	but at successively coarser levels of detail.
	//
	//		its(Atom/Bond)MeshSet[0]	coarsest
	//		its(Atom/Bond)MeshSet[1]
	//		its(Atom/Bond)MeshSet[2]
	//		its(Atom/Bond)MeshSet[3]	finest
	//
	var itsAtomsMeshSet
		= [GeometryGamesMeshBuffers<CrystalFlightUnlitVertex, CrystalFlightMeshIndexType>]()
	var itsBondsMeshSet
		= [GeometryGamesMeshBuffers<CrystalFlightUnlitVertex, CrystalFlightMeshIndexType>]()
	
	//	itsBoundingSphere encloses all the atoms in the unit cell
	//	and the endpoints of all the bonds in the unit cell.
	//
	//		Technical note:  It's possible that some tiny part
	//		of tube that represents a bond could extend
	//		slightly beyond that itsBoundingSphere, but
	//		it doesn't matter because each of a tube's endpoints
	//		gets covered by an atom (either an atom from the same copy
	//		of the unit cell, or an atom from a neighboring copy).
	//
	//	itsBoundingSphere may be slightly larger than a truly
	//	smallest bounding sphere, because our algorithm for computing it
	//	favors speed over optimality.
	//
	var itsBoundingSphere: Sphere = Sphere(	//	init() will overwrite this value
										center: SIMD3<Double>.zero,
										radius: 0.0)

	//	gMainScreenSupportsP3 is accessible only on the MainActor,
	//	so to access it from this non-isolated code
	//	the caller must pass it in explicitly.
	let itsMainScreenSupportsP3: Bool

	init(
		device: MTLDevice,
		crystal: Crystal,
		atomicRadiusFactor: Double,
		mainScreenSupportsP3: Bool
	) {
	
		itsDevice = device
		itsPreparedCrystal = crystal
		itsPreparedAtomicRadiusFactor = atomicRadiusFactor
		itsMainScreenSupportsP3 = mainScreenSupportsP3

		makeUnitCellAtomsMeshSet()
		makeUnitCellBondsMeshSet()
		computeBoundingSphere()
	}

	// MARK: -
	// MARK: Atoms in unit cell

	mutating func makeUnitCellAtomsMeshSet() {

		itsAtomsMeshSet = ( 0 ..< gNumLevelsOfDetail).map() { refinementLevel in
			makeAtomsMesh(refinementLevel: refinementLevel)
		}
	}
	
	func makeAtomsMesh(
		refinementLevel: Int
	) -> GeometryGamesMeshBuffers<CrystalFlightUnlitVertex, CrystalFlightMeshIndexType> {
	
		let theUnitSphereMesh: GeometryGamesMesh<GeometryGamesPlainVertex, CrystalFlightMeshIndexType>
			= unitSphereMesh(
				baseShape: .icosahedron,
				refinementLevel: refinementLevel)

		let theAtoms = itsPreparedCrystal.atoms
		
		//	Use theUnitSphereMesh as the building block to construct
		//	a single mesh that realizes all the atoms in the unit cell.

		let theVertices = theAtoms.flatMap() { atom in
		
			let theRadius = Float32(itsPreparedAtomicRadiusFactor * atom.radius)
			let theCenter = SIMD3<Float32>(atom.center)
			let theColor = atom.element.color(p3Gamut: itsMainScreenSupportsP3)
			
			let theAdjustedVertices = theUnitSphereMesh.itsVertices.map() { v in
			
				CrystalFlightUnlitVertex(
					pos: theCenter + theRadius * v.pos,
					nor: v.nor,
					col: theColor)
			}
			
			return theAdjustedVertices
		}

		let indexOffsetPerAtom = theUnitSphereMesh.itsVertices.count
		let theIndices = (0 ..< theAtoms.count).flatMap() { atomIndex in
		
			let theIndexOffset = CrystalFlightMeshIndexType(atomIndex * indexOffsetPerAtom)
			
			let theAdjustedIndices = theUnitSphereMesh.itsIndices.map() { vertexIndex in
				vertexIndex + theIndexOffset
			}
			
			return theAdjustedIndices
		}
		
		let theAtomMesh = GeometryGamesMesh(
							vertices: theVertices,
							indices: theIndices)

		let theAtomMeshBuffers = theAtomMesh.makeBuffers(device: itsDevice)

		return theAtomMeshBuffers
	}


	// MARK: -
	// MARK: Bonds in unit cell

	mutating func makeUnitCellBondsMeshSet() {

		itsBondsMeshSet = ( 0 ..< gNumLevelsOfDetail).map() { refinementLevel in
			makeBondsMesh(refinementLevel: refinementLevel)
		}
	}

	func makeBondsMesh(
		refinementLevel: Int
	) -> GeometryGamesMeshBuffers<CrystalFlightUnlitVertex, CrystalFlightMeshIndexType> {
	
		let theUnitCylinderMesh: GeometryGamesMesh<GeometryGamesPlainVertex, CrystalFlightMeshIndexType>
			= unitCylinderMesh(
				style: .antiprism,
				refinementLevel: refinementLevel)

		let theBonds = itsPreparedCrystal.bonds
		
		//	Use theUnitCylinderMesh as the building block to construct
		//	a single mesh that realizes all the bonds in the unit cell.

		let theVertices = theBonds.flatMap() { bond in
		
			let theRadius = bond.type.radius
			let theCenter = 0.5 * (bond.start + bond.finish)
			let theBondVector = bond.finish - bond.start
			let theHalfLength  = 0.5 * length(theBondVector)
			let theBondDirection = normalize(theBondVector)	//	theBondVector ≠ 0
			let theRawTubeDirection = SIMD3<Double>(0.0, 0.0, 1.0)
			let theRotationAsQuaternion = simd_quatd(
											from: theRawTubeDirection,
											to: theBondDirection)
			let theRotation = simd_double3x3(theRotationAsQuaternion)
				//	Even though it's faster to multiply quaternions
				//	than to multiply 3×3 matrices,
				//	when applying a transformation to a vector,
				//	a 3×3 matrix is faster than a quaternion.
			
			let theColor = bond.type.color

			let theAdjustedVertices = theUnitCylinderMesh.itsVertices.map() { v in
			
				let theRawPosition = SIMD3<Double>(v.pos)
				let theScaledPosition = SIMD3<Double> (
					theRawPosition.x * theRadius,
					theRawPosition.y * theRadius,
					theRawPosition.z * theHalfLength
				)
				let theRotatedPosition = theRotation * theScaledPosition
				let theTranslatedPosition = theRotatedPosition + theCenter
				
				let theRawNormal = SIMD3<Double>(v.nor)
				let theRotatedNormal = theRotation * theRawNormal

				let theAdjustedVertex = CrystalFlightUnlitVertex(
					pos: SIMD3<Float32>(theTranslatedPosition),
					nor: SIMD3<Float16>(theRotatedNormal),
					col: theColor)
				
				return theAdjustedVertex
			}
			
			return theAdjustedVertices
		}

		let indexOffsetPerBond = theUnitCylinderMesh.itsVertices.count
		let theIndices = (0 ..< theBonds.count).flatMap() { bondIndex in
		
			let theIndexOffset = CrystalFlightMeshIndexType(bondIndex * indexOffsetPerBond)
			
			let theAdjustedIndices = theUnitCylinderMesh.itsIndices.map() { vertexIndex in
				vertexIndex + theIndexOffset
			}
			
			return theAdjustedIndices
		}

		let theBondMesh = GeometryGamesMesh(
							vertices: theVertices,
							indices: theIndices)

		let theBondMeshBuffers = theBondMesh.makeBuffers(device: itsDevice)

		return theBondMeshBuffers
	}


	// MARK: -
	// MARK: Bounding sphere

	mutating func computeBoundingSphere() {
	
		//	Our algorithm is based on Ritter's algorithm
		//
		//		https://en.wikipedia.org/wiki/Bounding_sphere#Ritter's_bounding_sphere
		//
		//	but modified to take into account the atomic radii.
		//	By contrast, we ignore the bonds' tube radius,
		//	because any tiny bit of a bond's tube that extends
		//	beyond the computed bounding sphere will ultimately
		//	get covered by an atom (either an atom from this copy
		//	of the unit cell, or an atom from another copy).
		//	Well, that should be the case for the bounding spheres
		//	for all of the crystals that CrystalFlightCrystal provides.
		//	For a counterexample, you'd need a really huge bounding sphere,
		//	with a bond lying almost tangent to the sphere, but
		//	with both endpoints just inside it, and the middle of the tube
		//	just outside it.
		
		//	Pre-fetch the Arrays of Atoms and Bonds.
		//	In the case of .quartz (and only in the case of .quartz)
		//	this will require a small bit of computation
		//	to triple the hard-coded Atoms and Bonds
		//	to fill a full unit cell.
		//
		let theAtoms = itsPreparedCrystal.atoms
		let theBonds = itsPreparedCrystal.bonds
		
		//	Merge the atoms and the bonds' endpoints onto a single list of Spheres.
		//	Treat the bonds' endpoints as Spheres of radius zero.
		var theSpheres: [Sphere] = []
		theSpheres.reserveCapacity( theAtoms.count  +  2 * theBonds.count )
		for theAtom in theAtoms {
			theSpheres.append(Sphere(
				center: theAtom.center,
				radius: itsPreparedAtomicRadiusFactor * theAtom.radius))
		}
		for theBond in theBonds {
			theSpheres.append(Sphere(
				center: theBond.start,
				radius: 0.0))
			theSpheres.append(Sphere(
				center: theBond.finish,
				radius: 0.0))
		}
		
		//	Terminology
		//
		//	Throughout this document, when we talk about the "outdistance"
		//	between two spheres, we mean the distance from the far side
		//	of one to the far side of the other.  In other words,
		//	we mean the diameter of the smallest bounding sphere
		//	that contains the two given spheres.
		
		//	Maximally "outdistant" spheres
		//
		//	To get started with Ritter's algorithm, we need to find
		//	two spheres at maximal (or almost maximal) outdistance
		//	from each other.  To find them, we have a choice of algorithms:
		//
		//		- An O(n²) algorithm will find two spheres
		//			at truly maximal outdistance from each other.
		//
		//		- An O(n) algorithm will find two spheres
		//			that should be pretty close to maximally
		//			outdistant from each other, but might not
		//			be the absolutely maximal pair.
		//
		//	For the simple crystals that CrystalFlightCrystal provides,
		//	the O(n) algorithm finds an absolutely maximal pair,
		//	while the O(n²) algorithm runs acceptably fast,
		//	so either would work fine for our immediate purposes.
		//	I'm leaving the O(n²) algorithm in place in case
		//	it gives better results with any crystals that we might
		//	add in the future.  It seems unlikely that for any reasonable
		//	unit cell, that the run time would be a problem.  After all,
		//	this code gets run only occassionally, while a maximally good
		//	bounding sphere provides at least a small performance boost
		//	on every frame.
		//
		let (theSphereA, theSphereB) = maximallyOutdistantSpheresExact(theSpheres)	//	O(n²) algorithm
//		let (theSphereA, theSphereB) = maximallyOutdistantSpheresFast(theSpheres)	//	O(n)  algorithm

		//	Imitating Ritter's algorithm, start with a sphere that
		//	just barely encloses theSphereA and theSphereB.
		//
		var theBoundingSphere = enclosingSphere(theSphereA, theSphereB)
		
		//	Now examine all the other spheres (redundantly
		//	considering theSphereA and theSphereB as well)...
		for theSphere in theSpheres {
		
			if theBoundingSphere.encloses(theSphere) {
			
				//	Great!  No changes needed here.
				
			} else {
			
				//	Replace the existing bounding sphere with a new bounding sphere
				//	that encloses theSphere as well as the existing bounding sphere.
				//	This new sphere isn't maximally small, but it's good enough
				//	and it's easy and fast to compute.
				theBoundingSphere = enclosingSphere(theBoundingSphere, theSphere)
			}
		}

		itsBoundingSphere = theBoundingSphere
	}
	
	private func maximallyOutdistantSpheresExact(
		_ spheres: [Sphere]
	) -> (Sphere, Sphere) {
	
		let theSphereCount = spheres.count
		
		precondition(
			theSphereCount > 0,
			"List of spheres is empty")
		if theSphereCount < 2 {
			return (spheres[0], spheres[0])
		}
	
		var theFurthestSphereA = spheres[0]	//	tentative values, to be overwritten below
		var theFurthestSphereB = spheres[0]
		
		var theGreatestOutdistance = 0.0
		
		for i in 0 ... (spheres.count - 2) {
		
			let theSphereA = spheres[i]
				
			for j in (i+1) ... (spheres.count - 1) {
			
				let theSphereB = spheres[j]
				
				let theOutdistance = outdistance(theSphereA, theSphereB)
								
				if theOutdistance > theGreatestOutdistance {
				
					theFurthestSphereA = theSphereA
					theFurthestSphereB = theSphereB
					
					theGreatestOutdistance = theOutdistance
				}
			}
		}
		
		return (theFurthestSphereA, theFurthestSphereB)
	}
	
	private func maximallyOutdistantSpheresFast(
		_ spheres: [Sphere]
	) -> (Sphere, Sphere) {

		precondition(
			spheres.count > 0,
			"List of spheres is empty")

		//	Pick theSphereA arbitrarily
		let theSphereA = spheres[0]	//	tentative values, to be overwritten below

		//	Let theSphereB be a sphere that's maximally outdistant from theSphereA
		var theSphereB = spheres[0]
		var theGreatestOutdistanceAB = 0.0
		for theSphere in spheres {
		
			let theOutdistance = outdistance(theSphere, theSphereA)
							
			if theOutdistance > theGreatestOutdistanceAB {
			
				theSphereB = theSphere
				theGreatestOutdistanceAB = theOutdistance
			}
		}

		//	Let theSphereC be a sphere that's maximally outdistant from theSphereB
		var theSphereC = spheres[0]
		var theGreatestOutdistanceBC = 0.0
		for theSphere in spheres {
		
			let theOutdistance = outdistance(theSphere, theSphereB)
							
			if theOutdistance > theGreatestOutdistanceBC {
			
				theSphereC = theSphere
				theGreatestOutdistanceBC = theOutdistance
			}
		}
		
		//	In most cases, theSphereB and theSphereC will be almost maximally far apart.
		return (theSphereB, theSphereC)
	}
	
	private func outdistance(
		_ sphereA: Sphere,
		_ sphereB: Sphere
	) -> Double {
		
		return length(sphereB.center - sphereA.center)
			 + sphereA.radius
			 + sphereB.radius
	}
	
	private func enclosingSphere(
		_ sphereA: Sphere,
		_ sphereB: Sphere
	) -> Sphere {

		//	Does one sphere enclose the other?
		
		if sphereA.encloses(sphereB) {
			return sphereA
		}
		
		if sphereB.encloses(sphereA) {
			return sphereB
		}
		
		//	Compute a vector from the center of the first sphere
		//	to the center of the second, and note its length.
		
		let v = sphereB.center - sphereA.center
		let d = length(v)
		
		//	If the distance between the centers is small,
		//	but neither sphere encloses the other,
		//	then we probably have two almost identical spheres
		//	that differ only by some small numerical error.
		//	Treat this as a special case, to avoid numerical worries
		//	in the code for the general case, where we'll
		//	work with the unit vector u = v/d.

		if d < 0.001 {	//	ångströms
		
			return Sphere(
				center: 0.5 * (sphereA.center + sphereB.center),
				radius: max(sphereA.radius, sphereB.radius) + 0.5*d
			)
		}
		
		//	At this point it will be helpful to draw yourself
		//	a picture of two spheres (or just two circles),
		//	with a larger sphere (or circle) just barely enclosing them.
		
		let u = normalize(v)	//	unit vector pointing
								//	 from the center of the first  sphere
								//	 to the center of the second sphere

		let p₁ = sphereA.center		//	point where the the enclosing sphere
			   - sphereA.radius * u	//	 will touch the first sphere
									
		let p₂ = sphereB.center		//	point where the the enclosing sphere
			   + sphereB.radius * u	//	 will touch the second sphere
			
		let c = 0.5 * (p₁ + p₂)			//	enclosing sphere's center
		let r = 0.5 * length(p₂ - p₁)	//	enclosing sphere's radius
		
		return Sphere(center: c, radius: r)
	}
}
