//	CrystalFlightModel.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import Foundation
import CoreMotion
#if os(iOS)
import UIKit // for handleDeviceMotion()
#endif


let gDefaultCrystal = Crystal.iceHexagonal

let gMinAtomicRadiusFactor = 0.25
let gMaxAtomicRadiusFactor = 1.00
let gDefaultAtomicRadiusFactor = 0.75

let gMinHorizonRadius =  4.0		//	ångströms
let gMaxHorizonRadius = 32.0		//	ångströms
let gDefaultHorizonRadius = 18.0	//	ångströms

let gMaxUserSpeed = 2.0		//	ångströms/second
let gDefaultUserSpeed = 1.0	//	ångströms/second


@Observable class CrystalFlightModel: GeometryGamesModel {

	//	Which Crystal?
	var itsCrystal: Crystal = gDefaultCrystal {
		didSet {
			itsUserPosition = itsCrystal.userStartingPosition
			itsUserOrientation = gQuaternionIdentity
		}
	}

	//	The user may wish to shrink all atoms' radii,
	//	to show the crystal's large-scale structure more clearly.
	var itsAtomicRadiusFactor: Double = gDefaultAtomicRadiusFactor
			//	∈ [0.0, 1.0] in principle,
			//	but in practice we limit it to
			//	[gMinAtomicRadiusFactor, gMaxAtomicRadiusFactor]

	//	Everything out to itsHorizonRadius should be visible,
	//	with the fogging reaching pure black just as it reaches itsHorizonRadius.
	var itsHorizonRadius: Double = gDefaultHorizonRadius	//	ångströms

	//	Draw the bonds?
	var itsBondsAreVisible: Bool = true

	//	Enable fog?
	var itsFogIsEnabled: Bool = true

	//	Where is the user in the virtual space?
	var itsUserPosition: SIMD3<Double> = gDefaultCrystal.userStartingPosition
	var itsUserOrientation: simd_quatd = gQuaternionIdentity
	var itsUserSpeed: Double = (
		(gGetScreenshotOrientations || gMakeAppStoreScreenshots || gMakeWebSiteScreenshot) ?
		0.0 :
		gDefaultUserSpeed	//	ångströms/second
	)

	//	When the user taps to pause the motion, remember
	//	the current user speed, so that we may restore it
	//	when the user later taps again to unpause.
	//
	//		Special case:  If the user taps to pause and
	//		then -- before unpausing -- manually sets a new speed,
	//		the SpeedSlider resets itsPrePauseUserSpeed to nil.
	//
	var itsPrePauseUserSpeed: Double?

	//	Remember the device's previous interface attitude,
	//	which we'll use to suppress random jitters.
	var itsPreviousInterfaceAttitude: simd_quatd?

	var itsGestureRecognizerDelegate = CrystalFlightGestureRecognizerDelegate()


	override func updateModel() -> UInt64 {

#if os(iOS)
		//	If gMotionManager is present, use it.
		//
		//		Note:  At launch, the Motion Manager will typically
		//		not yet have any attitude data available, and
		//		deviceMotion will return nil.  After a short while,
		//		the attitude data will become available,
		//		with no action required on our part.
		//
		if let theDeviceAttitude = gMotionManager.deviceMotion?.attitude {
			handleDeviceMotion(theDeviceAttitude.rotationMatrix)
		}
#endif

		if itsUserSpeed != 0.0 {

			//	Move the user forward
			let theRawIncrement = SIMD3<Double>( 0.0, 0.0, itsUserSpeed * gFramePeriod )
			let theRotatedIncrement = itsUserOrientation.act(theRawIncrement)
			itsUserPosition += theRotatedIncrement
			
			//	The future AR version of Crystal Flight
			//	will let the user move freely in the space,
			//	even walking from room to room if the AR system
			//	supports that.  As the user wanders, we'll
			//	move the tiling along with him/her, keeping it
			//	based at whatever cell the user is currently in.
			//
			//	By contrast, this iOS version of Crystal Flight
			//	keeps the user near the origin, by applying
			//	primitiveTranslations as appropriate.
			keepUserNearOrigin()

			changeCount += 1
		}
		
		//	The user may or may not have modified itsUserOrientation,
		//	but let's normalize it once per frame here, so we don't
		//	have to remember to normalize in the various places
		//	where it might get modified.
		itsUserOrientation = itsUserOrientation.normalized

		return changeCount
	}

#if os(iOS)

	func handleDeviceMotion(_ rotationMatrix: CMRotationMatrix) {
	
		//	Ignore device motion when evaluating potential screenshot placements,
		//	and when creating screenshots.
		if (gGetScreenshotOrientations
		 || gMakeWebSiteScreenshot
		 || gMakeAppStoreScreenshots
		) {
			return
		}
		
		let theZAxis = SIMD3<Double>(0.0, 0.0, 1.0)

		let theIdentity			= simd_quatd(angle:  0.0 * .pi, axis: theZAxis)
		let theQuarterTurnCCW	= simd_quatd(angle: +0.5 * .pi, axis: theZAxis)
		let theQuarterTurnCW	= simd_quatd(angle: -0.5 * .pi, axis: theZAxis)
		let theHalfTurn 		= simd_quatd(angle:  1.0 * .pi, axis: theZAxis)

		//	As illustrated at
		//
		//		https://nshipster.com/cmdevicemotion/
		//	and
		//		https://developer.apple.com/documentation/coremotion/getting-processed-device-motion-data
		//
		//	CMAttitude uses a coordinate frame attached to the device, namely
		//
		//		the x axis runs left-to-right through the device's left and right edges,
		//		the y axis runs bottom-to-top through the device's bottom and top edges, and
		//		the z axis runs upward from the back side up through the display,
		//
		//	and reports the position of that coordinate frame relative
		//	to a fixed reference frame. We use the default fixed reference frame,
		//	which corresponds to the device lying face-up with its x- and y-axes
		//	pointing in the (horizontal!) directions that best match
		//	the physical device's orientation at the moment when we called
		//	startDeviceMotionUpdates().
		//
		let theDeviceAttitude = simd_quatd(
			ConvertCMRotationMatrixToCrystalFlightMatrix(rotationMatrix))

		//	The following code lets the main view show the crystal
		//	the same way before and after a device rotation.
		//	Unfortunately I've found no way to prevent iOS from animating
		//	the view's transition from portrait to landscape or vice versa.
		//	I would have preferred to suppress that animation,
		//	so that the crystal would stay rock solid on the display.
		//	I also failed to find any way to restrict the iPad
		//	to portrait-only (the check boxes in the info.plist file
		//	seem to have no effect).  So for now we just need to live
		//	with the ugly transition, and be glad that at least
		//	the crystal looks the same before and after.
		//
		let theInterfaceAttitude: simd_quatd
		if let theWindowScene = UIApplication.shared.connectedScenes.first as? UIWindowScene {

			let theInterfaceRotationFactor: simd_quatd
			switch theWindowScene.interfaceOrientation {
			case .unknown:
				theInterfaceRotationFactor = theIdentity
			case .portrait:
				theInterfaceRotationFactor = theIdentity
			case .portraitUpsideDown:
				theInterfaceRotationFactor = theHalfTurn
			case .landscapeLeft:
				theInterfaceRotationFactor = theQuarterTurnCCW
			case .landscapeRight:
				theInterfaceRotationFactor = theQuarterTurnCW
			@unknown default:
				theInterfaceRotationFactor = theIdentity
			}

			theInterfaceAttitude = theDeviceAttitude * theInterfaceRotationFactor
		}
		else {	//	should never occur
			theInterfaceAttitude = theDeviceAttitude
		}

		//	The first time this method gets called,
		//	itsPreviousInterfaceAttitude will be nil,
		//	but thereafter it will always be non-nil.
		if let thePreviousInterfaceAttitude = itsPreviousInterfaceAttitude {

			//	We want to distinguish purposeful device motions
			//	from random vibrations that the device might sense
			//	while sitting "motionless" on a table. What should
			//	the minimum change in the pitch, yaw or roll angle
			//	be for a motion to count as purposeful?
			//
			//		Note:  The threshold of 0.001 is fairly aggressive.
			//		It pretty much stops the animation when the device
			//		is lying still on a table and the scene isn't changing.
			//		The disadvantage is that if the user purposefully
			//		rotates the device very slowly (while the scene
			//		isn't otherwise changing) half or more of the frames
			//		won't get drawn.  In practice that's not a problem,
			//		though, because a suppressed frame's pitch, yaw and roll
			//		always (by the very nature of the algorithm) differ
			//		from the most recently drawn frame's pitch, yaw and roll
			//		by less than 0.001, which is hardly perceptable to the human eye.
			//
			let theIncrementalPitchYawRollThreshold = 0.001

			let theIncrementalRotation = thePreviousInterfaceAttitude.inverse
									   * theInterfaceAttitude

			if fabs(theIncrementalRotation.angle) > theIncrementalPitchYawRollThreshold {
			
				//	Crystal Flight, as well as responding to device motion,
				//	also lets the user drag the scene with his/her finger.
//	Commented out to suppress "unreachable code" warning
//				let allowsDragging = true
//				if !allowsDragging {
//
//					//	If we didn't allow dragging, itsUserOrientation
//					//	would always exactly match theInterfaceAttitude.
//					itsUserOrientation = theInterfaceAttitude
//				}
//				else {

					//	The user may use his/her finger to drag the scene,
					//	thus changing the relationship between itsUserOrientation
					//	and theInterfaceAttitude.  So here we should apply
					//	theIncrementalRotation to itsUserOrientation, so
					//	that the user's viewpoint rotates the same way
					//	the device does, even though their orientations differ.
					//
					//		Note:  This incremental-rotation approach
					//		would more-or-less work in the non-dragging case
					//		as well, but with the risk that tiny numerical errors
					//		might accumulate, thus causing the crystal to "drift"
					//		relative to the device's true orientation.
					//		The drifting might be negligible, but the direct approach
					//		(see "if" block above) is completely safe.
					//
					itsUserOrientation = itsUserOrientation * theIncrementalRotation
//				}

				changeCount += 1

				itsPreviousInterfaceAttitude = theInterfaceAttitude
			}
		}
		else {
			itsPreviousInterfaceAttitude = theInterfaceAttitude
		}
	}

	func ConvertCMRotationMatrixToCrystalFlightMatrix(_ aSrc: CMRotationMatrix) -> simd_double3x3
	{
		//	theAttitudeMatrix apparently uses left-to-right matrix action with row-major order
		//	(or, equivalently, right-to-left matrix action with column-major order)
		//	the same as we do, so there's no need to take a transpose here.

		//	Relative to a device held in portrait orientation,
		//	[theAttitude rotationMatrix] uses the frame (right vector, up vector, back vector)
		//	(see illustration at https://nshipster.com/cmdevicemotion/ ).
		//	But we'd rather work with (right vector, up vector, forward vector).
		//	So let's change coordinates to flip the z axis:
		//
		//		( 1  0  0 ) ( m₀₀  m₀₁  m₀₂ ) ( 1  0  0 )
		//		( 0  1  0 ) ( m₁₀  m₁₁  m₁₂ ) ( 0  1  0 )
		//		( 0  0 -1 ) ( m₂₀  m₂₁  m₂₂ ) ( 0  0 -1 )
		//	  =
		//		( 1  0  0 ) ( m₀₀  m₀₁ -m₀₂ ) ( 1  0  0 )
		//		( 0  1  0 ) ( m₁₀  m₁₁ -m₁₂ ) ( 0  1  0 )
		//		( 0  0 -1 ) (-m₂₀ -m₂₁  m₂₂ ) ( 0  0 -1 )
		//

		let theDst = simd_double3x3(
			SIMD3<Double>( aSrc.m11,  aSrc.m12, -aSrc.m13),
			SIMD3<Double>( aSrc.m21,  aSrc.m22, -aSrc.m23),
			SIMD3<Double>(-aSrc.m31, -aSrc.m32,  aSrc.m33)
		)
		
		return theDst
	}

#endif // os(iOS)

	func keepUserNearOrigin() {
	
		//	The hexagonal ice crystal nicely illustrates
		//	a choice we face here.  If we considered only
		//	the primitiveTranslations
		//
		//		(        0.0,                    0.0,            theIceHexC )
		//		( +0.5 * theIceHexAB, 0.5 * root3 * theIceHexAB,    0.0     )
		//		( -0.5 * theIceHexAB, 0.5 * root3 * theIceHexAB,    0.0     )
		//
		//	the unit cell would be a rhombic prism, but if we also
		//	consider the sum of those last two translations, namely
		//
		//		(    theIceHexAB,                0.0,               0.0     )
		//
		//	then the unit cell becomes a hexgonal prism.
		//	The advantage of working with the hexagonal prism
		//	is that its outradius is smaller, so it keeps us
		//	closer to the origin, and thus when we go to construct
		//	the tiling, fewer candidate translations will suffice
		//	to guarantee that we're displaying all atoms and bonds
		//	out to the desired tiling radius.
		
		let p = itsCrystal.primitiveTranslations
		
		var theTranslations: [SIMD3<Double>] = [
			 p[0],
			 p[1],
			 p[2],
			-p[0],
			-p[1],
			-p[2]
		]
		
		for i in 0 ... 1 {
			for j in i ... 2 {
				theTranslations.append(-p[i] - p[j])
				theTranslations.append(-p[i] + p[j])
				theTranslations.append( p[i] - p[j])
				theTranslations.append( p[i] + p[j])
			}
		}
		
		//	Apply a translation only when it brings the user
		//	closer to the origin by an unambiguously non-zero amount,
		//	to avoid the risk of oscillating back and forth
		//	if the user happens to arrive at a border
		//	between two copies of the fundamental cell.
		let ε = 0.0001
		
		var theDistance = length(itsUserPosition)
		var theDistanceWasDecreased: Bool
		repeat {
			
			theDistanceWasDecreased = false
			
			for theTranslation in theTranslations {
			
				let theCandidate = itsUserPosition + theTranslation
				let theCandidateDistance = length(theCandidate)
				if theCandidateDistance < theDistance - ε {
					itsUserPosition = theCandidate
					theDistance = length(itsUserPosition)
					theDistanceWasDecreased = true
				}
			}
		
		} while theDistanceWasDecreased
	}
	

// MARK: -
// MARK: Screenshot preparation

	func printUserBodyPlacementToConsole() {
		print("   position: \(itsUserPosition)")
		print("orientation: \(itsUserOrientation)")
		print("-----------------------------")
	}
}
