//	CrystalFlightRenderer.swift
//
//	© 2025 by Jeff Weeks
//	See TermsOfUse.txt

import Metal


class CrystalFlightRenderer: GeometryGamesRenderer<CrystalFlightModel> {

	//	When deciding how many instances to draw at each level of detail,
	//	it's handy to assign each level of detail a different color,
	//	so we can see where any visible polyhedralness remains.
	let itsColorByLevelOfDetail: Bool = false

	//	Before calling the superclass's (GeometryGamesRenderer's) init()
	//	to set up itsDevice, itsColorPixelFormat, itsCommandQueue, etc.,
	//	and indeed even before calling any of our own methods to set up
	//	the pipeline states, the mesh set, etc., we must ensure that
	//	all our instances variables have values.  For that reason,
	//	we declare them to be optionals, which are automatically initialized
	//	to nil.  On the other hand, unwrapping such optionals quickly
	//	becomes tedious, and makes the code harder to read (especially
	//	in more complex apps).  So let's define them as
	//	implicitly unwrapped optionals.  I'm slightly uncomfortable
	//	using implicitly unwrapped optionals -- I hadn't intended
	//	to use them at all, for run-time safety -- but in this case
	//	all we need to do to avoid problems is make sure that
	//	our own init() function provides non-nil values for these
	//	instances variables before returning.

	var itsFogPipelineState: MTLRenderPipelineState!
	var itsFoglessPipelineState: MTLRenderPipelineState!
	var itsTileCullingPipelineState: MTLComputePipelineState!
	var itsTilingArrayCompactionPipelineState: MTLComputePipelineState!
	var itsTransformationAndLightingPipelineState: MTLComputePipelineState!

	var itsDepthStencilState: MTLDepthStencilState!
	
	var itsUnlitMeshSet: CrystalFlightUnlitMeshSet!	//	for the unit cell

	//	Each lit atoms vertex buffer will contain the vertices
	//	for the atoms in the unit cell at a single refinement level,
	//	lit according to the user's current orientation,
	//	and similarly for the lit bonds vertex buffers.
	var itsTransformedAndLitAtomsVertices: [MTLBuffer] = [] // setUpMeshes() sets non-empty Array.
	var itsTransformedAndLitBondsVertices: [MTLBuffer] = [] // setUpMeshes() sets non-empty Array.

	var itsTilingBufferPool: GeometryGamesBufferPool!
	var itsIndirectArgumentsBufferPool: GeometryGamesBufferPool!


	init?() {

		//	Specify clear-color values in linear extended-range sRGB.
		let theClearColor = MTLClearColor(
								red: 0.0,
								green: 0.0,
								blue: 0.0,
								alpha: 1.0)
		
		super.init(
			wantsMultisampling: true,
			wantsDepthBuffer: true,
			clearColor: theClearColor)

		if !(
			setUpPipelineStates()
		 && setUpDepthStencilState()
		 && setUpMeshes()
		 && setUpInflightBufferPools()
		) {
			assertionFailure("Unable to set up Metal resources")
			return nil
		}
	}

	func setUpPipelineStates() -> Bool {
	
		guard let theGPUFunctionLibrary = itsDevice.makeDefaultLibrary() else {
			assertionFailure("Failed to create default GPU function library")
			return false
		}
		
		itsFogPipelineState
			= makeRenderPipelineState(gpuFunctionLibrary: theGPUFunctionLibrary, fog: true)
		if itsFogPipelineState == nil {
			return false
		}
		
		itsFoglessPipelineState
			= makeRenderPipelineState(gpuFunctionLibrary: theGPUFunctionLibrary, fog: false)
		if itsFoglessPipelineState == nil {
			return false
		}
	
		itsTileCullingPipelineState
			= makeTileCullingState(gpuFunctionLibrary: theGPUFunctionLibrary)
		if itsTileCullingPipelineState == nil {
			return false
		}
		
		itsTilingArrayCompactionPipelineState
			= makeTilingArrayCompactionState(gpuFunctionLibrary: theGPUFunctionLibrary)
		if itsTilingArrayCompactionPipelineState == nil {
			return false
		}
	
		itsTransformationAndLightingPipelineState
			= makeTransformationAndLightingState(gpuFunctionLibrary: theGPUFunctionLibrary)
		if itsTransformationAndLightingPipelineState == nil {
			return false
		}

		return true
	}
	
	func makeRenderPipelineState(
		gpuFunctionLibrary: MTLLibrary,
		fog: Bool
	) -> MTLRenderPipelineState? {

		var theUseFogValue = fog
		let theCompileTimeConstants = MTLFunctionConstantValues()
		theCompileTimeConstants.setConstantValue(&theUseFogValue, type: .bool, withName: "gUseFog")
		
		let theGPUVertexFunction: MTLFunction
		do {
			theGPUVertexFunction = try gpuFunctionLibrary.makeFunction(	//	returns non-nil
				name: "CrystalFlightVertexFunction",
				constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure(error.localizedDescription)
			return nil
		}

		guard let theGPUFragmentFunction = gpuFunctionLibrary.makeFunction(
				name: "CrystalFlightFragmentFunction") else {
			assertionFailure("Failed to load CrystalFlightFragmentFunction")
			return nil
		}

		//	Create a MTLVertexDescriptor describing the vertex attributes
		//	that the GPU sends along to each vertex.  Each vertex attribute
		//	may be a 1-, 2-, 3- or 4-component vector.
		
		let theVertexDescriptor = MTLVertexDescriptor()
		
			//	Say where to find each attribute.

		guard let thePosOffset = MemoryLayout.offset(of: \CrystalFlightLitVertex.pos) else {
			assertionFailure("Failed to find offset of CrystalFlightUnlitVertex.pos")
			return nil
		}
		theVertexDescriptor.attributes[VertexAttributePosition].format = MTLVertexFormat.float3
		theVertexDescriptor.attributes[VertexAttributePosition].bufferIndex = BufferIndexVertexAttributes
		theVertexDescriptor.attributes[VertexAttributePosition].offset = thePosOffset

		guard let theColOffset = MemoryLayout.offset(of: \CrystalFlightLitVertex.col) else {
			assertionFailure("Failed to find offset of CrystalFlightLitVertex.col")
			return nil
		}
		theVertexDescriptor.attributes[VertexAttributeColor].format = MTLVertexFormat.half3
		theVertexDescriptor.attributes[VertexAttributeColor].bufferIndex = BufferIndexVertexAttributes
		theVertexDescriptor.attributes[VertexAttributeColor].offset = theColOffset

			//	Say how to step through each buffer.

		theVertexDescriptor.layouts[BufferIndexVertexAttributes].stepFunction = MTLVertexStepFunction.perVertex
		theVertexDescriptor.layouts[BufferIndexVertexAttributes].stride = MemoryLayout<CrystalFlightLitVertex>.stride

		//	Create a MTLRenderPipelineDescriptor.
		
		let thePipelineDescriptor = MTLRenderPipelineDescriptor()
		thePipelineDescriptor.label = "Crystal Flight render pipeline"
		thePipelineDescriptor.rasterSampleCount = itsSampleCount
		thePipelineDescriptor.vertexFunction = theGPUVertexFunction
		thePipelineDescriptor.fragmentFunction = theGPUFragmentFunction
		thePipelineDescriptor.vertexDescriptor = theVertexDescriptor
		thePipelineDescriptor.colorAttachments[0].pixelFormat = itsColorPixelFormat
		thePipelineDescriptor.depthAttachmentPixelFormat = itsDepthPixelFormat
		thePipelineDescriptor.stencilAttachmentPixelFormat = MTLPixelFormat.invalid

		do {
			let thePipelineState = try itsDevice.makeRenderPipelineState(descriptor: thePipelineDescriptor)	//	returns non-nil
			return thePipelineState
		} catch {
			assertionFailure("Couldn't create itsPipelineState: \(error)")
			return nil
		}
	}
	
	func makeTileCullingState(
		gpuFunctionLibrary: MTLLibrary
	) -> MTLComputePipelineState? {

		guard let theTileCullingComputeFunction
		= gpuFunctionLibrary.makeFunction(name: "CrystalFlightTileCullingFunction") else {
			assertionFailure("CrystalFlightTileCullingFunction not found")
			return nil
		}

		let theTileCullingPipelineState: MTLComputePipelineState
		do {
			theTileCullingPipelineState = try itsDevice.makeComputePipelineState(
				function: theTileCullingComputeFunction)
		} catch {
			assertionFailure("Couldn't create theTileCullingPipelineState: \(error)")
			return nil
		}
		
		return theTileCullingPipelineState
	}
	
	func makeTilingArrayCompactionState(
		gpuFunctionLibrary: MTLLibrary
	) -> MTLComputePipelineState? {

		guard let theTilingArrayCompactionComputeFunction
		= gpuFunctionLibrary.makeFunction(name: "CrystalFlightTilingArrayCompactionFunction") else {
			assertionFailure("CrystalFlightTilingArrayCompactionFunction not found")
			return nil
		}

		let theTilingArrayCompactionPipelineState: MTLComputePipelineState
		do {
			theTilingArrayCompactionPipelineState = try itsDevice.makeComputePipelineState(
				function: theTilingArrayCompactionComputeFunction)
		} catch {
			assertionFailure("Couldn't create theTilingArrayCompactionPipelineState: \(error)")
			return nil
		}
		
		return theTilingArrayCompactionPipelineState
	}
	
	func makeTransformationAndLightingState(
		gpuFunctionLibrary: MTLLibrary
	) -> MTLComputePipelineState? {

		var theMakeAppIconValue				= (gMakeIcon ? true : false)
		var theColorByLevelOfDetailValue	= (itsColorByLevelOfDetail ? true : false)

		let theCompileTimeConstants = MTLFunctionConstantValues()
		theCompileTimeConstants.setConstantValue(&theMakeAppIconValue,			type: .bool, withName: "gMakeAppIcon"			)
		theCompileTimeConstants.setConstantValue(&theColorByLevelOfDetailValue,	type: .bool, withName: "gColorByLevelOfDetail"	)

		let theTransformationAndLightingComputeFunction: MTLFunction
		do {
			theTransformationAndLightingComputeFunction
				= try gpuFunctionLibrary.makeFunction(	//	returns non-nil
					name: "CrystalFlightTransformationAndLightingFunction",
					constantValues: theCompileTimeConstants)
		} catch {
			assertionFailure(error.localizedDescription)
			return nil
		}
		
		let theTransformationAndLightingPipelineState: MTLComputePipelineState
		do {
			theTransformationAndLightingPipelineState = try itsDevice.makeComputePipelineState(
				function: theTransformationAndLightingComputeFunction)
		} catch {
			assertionFailure("Couldn't create theTransformationAndLightingPipelineState: \(error)")
			return nil
		}
		
		return theTransformationAndLightingPipelineState
	}
	
	func setUpDepthStencilState() -> Bool {

		//	init() has already checked that itsDepthPixelFormat != MTLPixelFormat.invalid
		
		let theDepthStencilDescriptor = MTLDepthStencilDescriptor()
		theDepthStencilDescriptor.depthCompareFunction = MTLCompareFunction.less
		theDepthStencilDescriptor.isDepthWriteEnabled = true

		itsDepthStencilState = itsDevice.makeDepthStencilState(
									descriptor: theDepthStencilDescriptor)
		if itsDepthStencilState == nil {
			assertionFailure("Couldn't create itsDepthStencilState")
			return false
		}

		return true
	}
	
	func setUpMeshes() -> Bool {
	
		//	Create an arbitrary CrystalFlightUnlitMeshSet as part of
		//	this CrystalFlightRenderer's init() sequence, so that
		//	itsUnlitMeshSet can be an implicitly unwrapped optional,
		//	which keeps subsequent code cleaner (compared to using
		//	a manually unwrapped optional).  At render time --
		//	when the ModelData will be available -- encodeCommands()
		//	will update this arbitrary CrystalFlightUnlitMeshSet
		//	to the desired crystalType and atomicRadiusFactor.
		
		let theArbitraryCrystal: Crystal = .iceHexagonal
		let theArbitraryAtomicRadiusFactor = gMaxAtomicRadiusFactor
		
		itsUnlitMeshSet = CrystalFlightUnlitMeshSet(
							device: itsDevice,
							crystal: theArbitraryCrystal,
							atomicRadiusFactor: theArbitraryAtomicRadiusFactor,
								//	gMainScreenSupportsP3 is accessible only
								//	on the MainActor, so we need to pass it
								//	into the CrystalFlightUnlitMeshSet.
							mainScreenSupportsP3: gMainScreenSupportsP3)
		
		//	Create vertex buffers of some arbitrary size for
		//	the transformed and lit vertices. encodeCommands()
		//	will replace them with buffers of the required size
		//	(which may vary as the user changes the crystal).
		//
		func makeDefaultBuffers() -> [MTLBuffer] {
		
			let theBuffers = (0 ..< gNumLevelsOfDetail).map() { _ in
			
				guard let theBuffer = itsDevice.makeBuffer(
					length: 1 * MemoryLayout<CrystalFlightLitVertex>.stride,
					options: .storageModePrivate)
				else { preconditionFailure("setUpMeshes() failed to create theBuffer") }
				
				return theBuffer
			}
			
			return theBuffers
		}
		itsTransformedAndLitAtomsVertices = makeDefaultBuffers()
		itsTransformedAndLitBondsVertices = makeDefaultBuffers()

		return true
	}
	
	func setUpInflightBufferPools() -> Bool {

		//	Each tiling buffer will contain gNumLevelsOfDetail contiguous arrays
		//	of CrystalFlightTiles, one array for each level of detail.
		let theOriginalTileCount = 1
		itsTilingBufferPool = GeometryGamesBufferPool(
			device: itsDevice,
			initialBufferSize: gNumLevelsOfDetail * theOriginalTileCount * MemoryLayout<CrystalFlightTile>.stride,
			storageMode: .storageModePrivate,
			bufferLabel: "Tiling buffer")
					
		//	Each indirect arguments buffer will contain an array
		//	of 2*gNumLevelsOfDetail instances of MTLDrawIndexedPrimitivesIndirectArguments.
		//	The first gNumLevelsOfDetail instances will be used to draw the atoms,
		//	while the second gNumLevelsOfDetail instances will be used to draw the bonds.
		itsIndirectArgumentsBufferPool = GeometryGamesBufferPool(
			device: itsDevice,
			initialBufferSize: 2 * gNumLevelsOfDetail *
				MemoryLayout<MTLDrawIndexedPrimitivesIndirectArguments>.stride,
			storageMode: .storageModeShared,
			bufferLabel: "Indirect arguments buffer")
		
		return true
	}


// MARK: -
// MARK: Rendering

	//	GeometryGamesRenderer (our superclass) provides the functions
	//
	//		render()
	//			for standard onscreen rendering
	//
	//		createOffscreenImage()
	//			for CopyImage and SaveImage
	//
	//	Those two functions handle the app-independent parts
	//	of rendering an image, but for the app-dependent parts
	//	they call encodeCommands(), which we override here
	//	to provide the app-specific content.

	override func encodeCommands(
		modelData: CrystalFlightModel,
		commandBuffer: MTLCommandBuffer,
		renderPassDescriptor: MTLRenderPassDescriptor,
		frameWidth: Int,
		frameHeight: Int,
		transparentBackground: Bool,	//	Only KaleidoPaint and KaleidoTile need to know
										//	transparentBackground while encoding commands.
		extraRenderFlag: Bool?,
		quality: GeometryGamesImageQuality
	) {

		//	Note:  If CrystalFlight someday supports only A11 GPUs and later,
		//	I can delete theNonuniformThreadGroupSizesAreAvailable along
		//	with the fallback code that it invokes.
		//
		let theNonuniformThreadGroupSizesAreAvailable = itsDevice.supportsFamily(.apple4)	//	A11 GPU and later

		//	Let itsUnlitMeshSet update its mesh, if necessary,
		//	to match any recent changes to itsCrystal or itsAtomicRadiusFactor.
		itsUnlitMeshSet.itsPreparedCrystal = modelData.itsCrystal
		itsUnlitMeshSet.itsPreparedAtomicRadiusFactor = modelData.itsAtomicRadiusFactor

		var theUniformData = prepareUniformData(
								modelData: modelData,
								frameWidth: frameWidth,
								frameHeight: frameHeight)

		//*********************************************
		//
		//	Transform & light the meshes
		//
		//*********************************************
		
		//	Submit a MTLComputeCommandEncoder to transform
		//	the meshes' vertices from world space to camera space
		//	and to light them.  It's more efficient to let
		//	the MTLComputeCommandEncoder do this once
		//	instead of asking the vertex shader to do it over and over,
		//	once for every visible instance of the unit cell.
		
		guard let theTnLComputeEncoder = commandBuffer.makeComputeCommandEncoder() else
		{
			assertionFailure("encodeCommands() failed to create theRotationAndLightingComputeEncoder")
			return
		}
		theTnLComputeEncoder.label = "Rotate and light vertices"
		theTnLComputeEncoder.setComputePipelineState(itsTransformationAndLightingPipelineState)
		
		var theInverseRotation: simd_float3x3
			= convertDouble3x3toFloat3x3(simd_double3x3(modelData.itsUserOrientation.inverse))
		theTnLComputeEncoder.setBytes(
			&theInverseRotation,
			length: MemoryLayout<simd_float3x3>.stride,
			index: BufferIndexRotation)

		var theInverseTranslation: SIMD3<Float32>
			= SIMD3<Float32>(-modelData.itsUserPosition)
		theTnLComputeEncoder.setBytes(
			&theInverseTranslation,
			length: MemoryLayout<SIMD3<Float32>>.stride,
			index: BufferIndexTranslation)
		
		//	Rotate and light the atom vertices for each level of detail.
		transformAndLightOneSet(
			computeEncoder: theTnLComputeEncoder,
			unlitSet: itsUnlitMeshSet.itsAtomsMeshSet,
			litVertices: &itsTransformedAndLitAtomsVertices)
													
		//	Rotate and light the bond vertices for each level of detail.
		transformAndLightOneSet(
			computeEncoder: theTnLComputeEncoder,
			unlitSet: itsUnlitMeshSet.itsBondsMeshSet,
			litVertices: &itsTransformedAndLitBondsVertices)

		theTnLComputeEncoder.endEncoding()


		//*********************************************
		//
		//	Tiling and frustum culling
		//
		//*********************************************

		let thePrimitiveTranslationsInCameraCoordinates
			= primitiveTranslationsInCameraCoordinates(modelData: modelData)
		
		let theViewFrustumOutwardUnitNormals
			= viewFrustumOutwardUnitNormals(
				frameWidth: frameWidth,
				frameHeight: frameHeight)
		
		let theBoundingSphereInCameraCoordinates
			= Sphere(
				center: modelData.itsUserOrientation.inverse.act(
					itsUnlitMeshSet.itsBoundingSphere.center - modelData.itsUserPosition),
				radius: itsUnlitMeshSet.itsBoundingSphere.radius)

		let theLatticeBounds: [ClosedRange<Int>]
			= latticeBounds(
				modelData: modelData,
				primitiveTranslationsInCameraCoordinates:
					thePrimitiveTranslationsInCameraCoordinates,
				boundingSphereInCameraCoordinates:
					theBoundingSphereInCameraCoordinates)

		let theDimensions: [Int] = theLatticeBounds.map({ bounds in
			bounds.upperBound - bounds.lowerBound + 1})
			
		let theRawTileCount: Int = theDimensions.reduce(1,
			{runningProduct, dimension in runningProduct * dimension})
		
		//	At this point we need to think ahead.  After we submit
		//	theTileCullingComputeEncoder, the next step will be to run
		//	theTilingArrayCompactionComputeEncoder, which requires
		//	a power-of-two threadgroup height.
		//	So let's choose that threadgroup height now...
		//
		let theMaxCompactionThreadgroupHeight = min(
			Int(gCompactionThreadgroupMaxHeight),	//	cast from unsigned C-style to Int
				//	size of staticly allocated buffer in threadgroup memory
			itsTilingArrayCompactionPipelineState.maxTotalThreadsPerThreadgroup)
		var theTenativeCompactionThreadgroupHeight = itsTilingArrayCompactionPipelineState.threadExecutionWidth
		while 2 * theTenativeCompactionThreadgroupHeight <= theMaxCompactionThreadgroupHeight {
			theTenativeCompactionThreadgroupHeight *= 2
		}
		let theCompactionThreadgroupHeight = theTenativeCompactionThreadgroupHeight
		//
		//	...and pad theTileCount to be an integer multiple of it.
		//	The padding will contain unused tiles, which behave
		//	just like culled tiles.
		//
		let theCompactionLoopCount
			= (theRawTileCount + (theCompactionThreadgroupHeight - 1))	//	round up
			/ theCompactionThreadgroupHeight
		let theTileCount = theCompactionLoopCount * theCompactionThreadgroupHeight

		let theTileByteCount = MemoryLayout<CrystalFlightTile>.stride
		let theBufferByteCount = gNumLevelsOfDetail * theTileCount * theTileByteCount
		itsTilingBufferPool.ensureSize(theBufferByteCount)
		
		let theTilingBuffer = itsTilingBufferPool.get(forUseWith: commandBuffer)

		//	Submit a MTLBlitCommandEncoder to clear theTilingBuffer.
		//	In particular, we want to initialize the fourth coordinate
		//	of each CrystalFlightTile to 0.0.
		//
		guard let theBlitEncoder = commandBuffer.makeBlitCommandEncoder() else
		{
			assertionFailure("encodeCommands() failed to create theBlitEncoder")
			return
		}
		theBlitEncoder.label = "clear tiling buffer"
		theBlitEncoder.fill(
			buffer: theTilingBuffer,
			range: 0 ..< theBufferByteCount,
			value: 0)
		theBlitEncoder.endEncoding()

		//	Submit a MTLComputeCommandEncoder to write all tiles
		//	to theTilingBuffer.  The GPU will be testing lots
		//	of different tiles in parallel, so we'll write each CrystalFlightTile
		//	to a pre-determined location in theTilingBuffer.
		//	See the comment accompanying the CrystalFlightTile typedef
		//	for an explanation of each tile's fourth coordinate.
		
		guard let theTileCullingComputeEncoder = commandBuffer.makeComputeCommandEncoder() else
		{
			assertionFailure("encodeCommands() failed to create theTileCullingComputeEncoder")
			return
		}
		theTileCullingComputeEncoder.label = "Tile culling"
		theTileCullingComputeEncoder.setComputePipelineState(itsTileCullingPipelineState)

		var theTilingParameters = CrystalFlightTilingParameters(
			itsTileCount: UInt32(theTileCount),
			itsGridLowerLimits: SIMD3<Int16>(
									Int16(theLatticeBounds[0].lowerBound),
									Int16(theLatticeBounds[1].lowerBound),
									Int16(theLatticeBounds[2].lowerBound)),
			itsGridDimensions: SIMD3<UInt16>(
									UInt16(theDimensions[0]),
									UInt16(theDimensions[1]),
									UInt16(theDimensions[2])),
			itsGridTranslations: convertDouble3x3toFloat3x3(thePrimitiveTranslationsInCameraCoordinates),
			itsOutwardUnitNormals: convertDouble4x3toFloat4x3(theViewFrustumOutwardUnitNormals),
			itsBoundingSphereCenter: SIMD3<Float32>(theBoundingSphereInCameraCoordinates.center),
			itsBoundingSphereRadius: Float32(theBoundingSphereInCameraCoordinates.radius),
			itsAdjustedHorizonRadius: Float32(modelData.itsHorizonRadius
										+ theBoundingSphereInCameraCoordinates.radius),
			itsDistanceCutoffs: (
				(gMakeAppStoreScreenshots || gMakeWebSiteScreenshot) ?
				SIMD3<Float16>(
					Float16.greatestFiniteMagnitude,	//	best quality for screenshots
					Float16.greatestFiniteMagnitude,
					Float16.greatestFiniteMagnitude
				) :
				SIMD3<Float16>(
					(modelData.itsFogIsEnabled ? 24.0 : 32.0),
					12.0,
					6.0
				)
			)
		)
		theTileCullingComputeEncoder.setBytes(
			&theTilingParameters,
			length: MemoryLayout<CrystalFlightTilingParameters>.stride,
			index: BufferIndexTilingParameters)

		theTileCullingComputeEncoder.setBuffer(
			theTilingBuffer,
			offset: 0,
			index: BufferIndexCulledTiling)

	
		let theTotalThreads = MTLSizeMake(
								theDimensions[0],
								theDimensions[1],
								theDimensions[2])
		if theNonuniformThreadGroupSizesAreAvailable {

			//	maxTotalThreadsPerThreadgroup can never exceed
			//
			//		 512 for devices in MTLGPUFamilyApple3 and older (A10 SoC and older)
			//		1024 for devices in MTLGPUFamilyApple4 and newer (A11 SoC and newer)
			//
			//	In principle it could be less in cases where a compute function
			//	has heavy register needs, but itsTileCullingPipelineState is simple enough
			//	that it can operate at those maximum levels.
			//
			//	Putting the whole grid into a single threadgroup, while possible
			//	for smaller grids like those we find for hexagonal ice and quartz
			//	with a moderate horizon radius, would be counterproductive, because
			//	(if my understanding of threadgroups is correct) this would
			//	leave all the threads sharing time on a single GPU core
			//	while the other GPU cores all sit idle.
			//
			//	Instead, our strategy will be to divide each 2D plane of the grid
			//	into as few threadgroups as possible.  Typically a single threadgroup
			//	suffices, although for a diamond crystal with a large horizon radius,
			//	two threadgroups are required even on an A11 device (the grid has
			//	dimensions 33×33×33 and 33*33 = 1089 > 1024).
			
			//	Note:  While our first priority is to divide each 2D plane
			//	into as few threadgroups as possible, our second priority is
			//	to make those threadgroups as small as possible.
			//	For example, we'd rather divide a 33×33 plane
			//	into four 17×17 threadgroups that are almost fully occupied,
			//	rather than four 32×32 threadgroups, three of which are
			//	almost entirely empty.  (To be honest, though, I don't know
			//	enough about GPUs to say with conviction that the 17×17 threadgroups
			//	would be more efficient, but it seems plausible.)

			let theMaxRowsPerThreadgroup	//	round down
				= itsTileCullingPipelineState.maxTotalThreadsPerThreadgroup
				/ theTotalThreads.depth
				
			precondition(
				theMaxRowsPerThreadgroup > 0,
				"Grid size is larger than maxTotalThreadsPerThreadgroup")
			
			let theMinThreadgroupsPerPlane	//	round up
				= ( theTotalThreads.height + (theMaxRowsPerThreadgroup - 1) )
				/ theMaxRowsPerThreadgroup

			let theRowsPerThreadgroup		//	round up
				= ( theTotalThreads.height + (theMinThreadgroupsPerPlane - 1) )
				/ theMinThreadgroupsPerPlane
			
			let theThreadsPerThreadgroup
				= MTLSizeMake(
					1,
					theRowsPerThreadgroup,
					theTotalThreads.depth)
			
			theTileCullingComputeEncoder.dispatchThreads(
				theTotalThreads,
				threadsPerThreadgroup: theThreadsPerThreadgroup)

		} else {
		
			//	Non-uniform threadgroup sizes aren't available,
			//	so let's take a simpler approach that's less efficient
			//	but completely safe.
			let theThreadsPerThreadgroup: MTLSize
			if theDimensions[1] * theDimensions[2]	//	rarely fails
				<= itsTileCullingPipelineState.maxTotalThreadsPerThreadgroup {
			
				theThreadsPerThreadgroup = MTLSizeMake(1, theDimensions[1], theDimensions[2])
									
			} else if theDimensions[2]	//	hard to imagine that this would fail for any reasonable grid
				<= itsTileCullingPipelineState.maxTotalThreadsPerThreadgroup {

				theThreadsPerThreadgroup = MTLSizeMake(1, 1, theDimensions[2])

			} else {	//	the ultimate failsafe fallback

				theThreadsPerThreadgroup = MTLSizeMake(1, 1, 1)

			}

			let theTotalThreadgroups = MTLSizeMake(
								theTotalThreads.width  / theThreadsPerThreadgroup.width,
								theTotalThreads.height / theThreadsPerThreadgroup.height,
								theTotalThreads.depth  / theThreadsPerThreadgroup.depth)
			
			theTileCullingComputeEncoder.dispatchThreadgroups(
				theTotalThreadgroups,
				threadsPerThreadgroup: theThreadsPerThreadgroup)
		}

		theTileCullingComputeEncoder.endEncoding()


		//*********************************************
		//
		//	Compact the tiling arrays and prepare indirect argument buffers
		//
		//*********************************************

		guard let theTilingArrayCompactionComputeEncoder = commandBuffer.makeComputeCommandEncoder() else
		{
			assertionFailure("encodeCommands() failed to create theTilingArrayCompactionComputeEncoder")
			return
		}
		theTilingArrayCompactionComputeEncoder.label = "Tiling array compaction"
		theTilingArrayCompactionComputeEncoder.setComputePipelineState(itsTilingArrayCompactionPipelineState)
	
		theTilingArrayCompactionComputeEncoder.setBuffer(
			theTilingBuffer,
			offset: 0,
			index: BufferIndexCompactableTiling)

		var theLoopCount16 = UInt16(theCompactionLoopCount)
		theTilingArrayCompactionComputeEncoder.setBytes(
			&theLoopCount16,
			length: MemoryLayout<UInt16>.stride,
			index: BufferIndexLoopCount)

		let theIndirectArgumentsBuffer = itsIndirectArgumentsBufferPool.get(forUseWith: commandBuffer)
		let theIndirectArguments = theIndirectArgumentsBuffer.contents().bindMemory(
									to: MTLDrawIndexedPrimitivesIndirectArguments.self,
									capacity: 2 * gNumLevelsOfDetail)
		for i in 0 ..< gNumLevelsOfDetail {
		
			//	atoms at level of detail i
			theIndirectArguments[i].indexStart = 0
			theIndirectArguments[i].indexCount = UInt32(itsUnlitMeshSet.itsAtomsMeshSet[i].indexCount)
			theIndirectArguments[i].baseVertex = 0
			theIndirectArguments[i].baseInstance = UInt32(i) * UInt32(theTileCount)
			theIndirectArguments[i].instanceCount = 0	//	compute encoder will set this

			//	bonds at level of detail i
			let j = gNumLevelsOfDetail + i
			theIndirectArguments[j].indexStart = 0
			theIndirectArguments[j].indexCount = UInt32(itsUnlitMeshSet.itsBondsMeshSet[i].indexCount)
			theIndirectArguments[j].baseVertex = 0
			theIndirectArguments[j].baseInstance = UInt32(i) * UInt32(theTileCount)
			theIndirectArguments[j].instanceCount = 0	//	compute encoder will set this
		}
		theTilingArrayCompactionComputeEncoder.setBuffer(
			theIndirectArgumentsBuffer,
			offset: 0,
			index: BufferIndexIndirectArguments)

		if theNonuniformThreadGroupSizesAreAvailable {

			theTilingArrayCompactionComputeEncoder.dispatchThreads(
				MTLSizeMake(gNumLevelsOfDetail, theCompactionThreadgroupHeight, 1),
				threadsPerThreadgroup: MTLSizeMake(1, theCompactionThreadgroupHeight, 1))
		} else {
		
			theTilingArrayCompactionComputeEncoder.dispatchThreadgroups(
				MTLSizeMake(gNumLevelsOfDetail, 1, 1),
				threadsPerThreadgroup: MTLSizeMake(1, theCompactionThreadgroupHeight, 1))
		}

		theTilingArrayCompactionComputeEncoder.endEncoding()

		//*********************************************
		//
		//	Render the tiling using theIndirectArgumentsBuffer
		//
		//*********************************************
		
		guard let theRenderEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: renderPassDescriptor) else
		{
			assertionFailure("encodeCommands() failed to create theRenderEncoder")
			return
		}
		theRenderEncoder.label = "Crystal Flight render encoder"
		theRenderEncoder.setRenderPipelineState(modelData.itsFogIsEnabled ?
									itsFogPipelineState : itsFoglessPipelineState)
		theRenderEncoder.setDepthStencilState(itsDepthStencilState)
		theRenderEncoder.setCullMode(MTLCullMode.back)
		theRenderEncoder.setFrontFacing(MTLWinding.clockwise)	//	the default
		theRenderEncoder.setVertexBytes(
							&theUniformData,
							length: MemoryLayout<CrystalFlightUniformData>.stride,
							index: BufferIndexUniforms)
		theRenderEncoder.setVertexBuffer(theTilingBuffer, offset: 0, index: BufferIndexTiling)

		for theLevel in 0 ..< gNumLevelsOfDetail {
		
			//	Draw atoms (always)
			do {
				//	Use theTransformedAndLitAtomsVertices along with
				//	the unmodified index data from theAtomsMesh.
				let theAtomsMesh = itsUnlitMeshSet.itsAtomsMeshSet[theLevel]
				let theTransformedAndLitAtomsVertices = itsTransformedAndLitAtomsVertices[theLevel]
		
				theRenderEncoder.setVertexBuffer(
									theTransformedAndLitAtomsVertices,
									offset: 0,
									index: BufferIndexVertexAttributes)

				theRenderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexType: theAtomsMesh.mtlIndexType,
								indexBuffer: theAtomsMesh.indexBuffer,
								indexBufferOffset: 0,
								indirectBuffer: theIndirectArgumentsBuffer,
								indirectBufferOffset:
									(0 + theLevel) *
									MemoryLayout<MTLDrawIndexedPrimitivesIndirectArguments>.stride)
			}
			
			//	Draw bonds if desired
			if modelData.itsBondsAreVisible {
				
				//	Use theTransformedAndLitBondsVertices along with
				//	the unmodified index data from theBondsMesh.
				let theBondsMesh = itsUnlitMeshSet.itsBondsMeshSet[theLevel]
				let theTransformedAndLitBondsVertices = itsTransformedAndLitBondsVertices[theLevel]
			
				theRenderEncoder.setVertexBuffer(
									theTransformedAndLitBondsVertices,
									offset: 0,
									index: BufferIndexVertexAttributes)

				theRenderEncoder.drawIndexedPrimitives(
								type: .triangle,
								indexType: theBondsMesh.mtlIndexType,
								indexBuffer: theBondsMesh.indexBuffer,
								indexBufferOffset: 0,
								indirectBuffer: theIndirectArgumentsBuffer,
								indirectBufferOffset:
									(gNumLevelsOfDetail + theLevel) *
									MemoryLayout<MTLDrawIndexedPrimitivesIndirectArguments>.stride)
			}
		}

		theRenderEncoder.endEncoding()
	}
		
	func prepareUniformData(
		modelData: CrystalFlightModel,
		frameWidth: Int,
		frameHeight: Int
	) -> CrystalFlightUniformData {
		
		let theProjectionMatrix = makeProjectionMatrix(
									frameWidth: Double(frameWidth),
									frameHeight: Double(frameHeight))

		let theHorizonRadius = Float16(modelData.itsHorizonRadius)	//	in ångströms

		let theFogBegin = 0.5 * modelData.itsHorizonRadius	//	in ångströms
		let theFogRangeInverse =							//	in ångströms⁻¹
			1.0 / (modelData.itsHorizonRadius - theFogBegin)

		let theUniformData = CrystalFlightUniformData(
			itsProjectionMatrix: convertDouble4x4toFloat4x4(theProjectionMatrix),
			itsHorizonRadius: Float16(theHorizonRadius),
			itsFogBegin: Float16(theFogBegin),
			itsFogRangeInverse: Float16(theFogRangeInverse))

		return theUniformData
	}
	
	func makeProjectionMatrix(
		frameWidth: Double,	//	typically in pixels or points
		frameHeight: Double	//	typically in pixels or points
	) -> simd_double4x4 {	//	returns the projection matrix
	
		if frameWidth <= 0.0 || frameHeight <= 0.0 {
		
			assertionFailure("nonpositive-size image received")
			return matrix_identity_double4x4
		}
		
		let w = 0.5 * frameWidth	//	view half-width  (typically pixels or points)
		let h = 0.5 * frameHeight	//	view half-height (typically pixels or points)
		if w <= 0.0 || h <= 0.0 {
			assertionFailure("nonpositive-size image received")
			return matrix_identity_double4x4
		}
		
		//	The characteristic view size is the distance in the view,
		//	measured from the center of the view outwards,
		//	that subtends a 45° angle in the observer's field-of-view.
		//	Note that it's measured in pixels or points, which is OK
		//	because all we really care about are the ratios w/c and h/c.
		let c = characteristicViewSize(
					frameWidth: frameWidth,
					frameHeight: frameHeight)
		if c <= 0.0 {
			assertionFailure("nonpositive characteristic size")
			return matrix_identity_double4x4
		}
		
		//	Let n and f be the distance from the observer at (0,0,0,1)
		//	to the near clipping plane z = n and the far clipping plane z = f, respectively.
		//
		//		Note:  theNearClipDistance shouldn't be too small,
		//		to avoid needless loss of depth buffer precision.
		//
		let theNearClipDistance = 0.125	//	in ångströms
		let n = theNearClipDistance		//	A shorter name for the same thing.
	//	let f → ∞						//	We may conceptually set f = ∞.


		////////////////////////////////////
		//	View wedge
		////////////////////////////////////

		//	If the view were square, the rays through the points
		//
		//		(±n, ±n, n, 1)
		//	and
		//		(±f, ±f, f, 1)
		//
		//	would define a view frustum of angular width and angular height
		//	both equal to ±45°, as seen by an observer at (0,0,0,1).
		//	Typically, though, the view isn't square, but instead has
		//	unequal half-width w and half-height h.
		//
		//	For this not-necessarily-square view, the rays through the points
		//
		//		(±n*(w/c), ±n*(h/c), n, 1)
		//	and
		//		(±f*(w/c), ±f*(h/c), f, 1)
		//
		//	define a view frustum that subtends the correct angle
		//	in each direction, relative to an observer at the ray through (0,0,0,1).
		//
		//	It's convenient to project those points
		//	from the hyperplane w = 1 onto the hyperplane z = 1, giving
		//
		//		(±w/c, ±h/c, 1, 1/n)
		//	and
		//		(±w/c, ±h/c, 1, 1/f)
		//
		//	because those projected points are the corners of a rectangular box.
		//	The rays defined by those points remain the same,
		//	and the observer remains at the ray through (0,0,0,1).
		//
		
		////////////////////////////////////
		//	Clipping wedge
		////////////////////////////////////

		//	Metal clips to the wedge
		//
		//		-w ≤ x ≤ w
		//		-w ≤ y ≤ w
		//		 0 ≤ z ≤ w
		//
		//	The intersection of this wedge with the hyperplane at w = 1
		//	is a box with vertices at
		//
		//		(±1, ±1, 0, 1)
		//	and
		//		(±1, ±1, 1, 1)
		//
		//	The observer sits on the ray through (0,0,-1,0).
		
		////////////////////////////////////
		//	Projection matrix
		////////////////////////////////////

		//	We want to construct a projection matrix that takes the view wedge
		//	whose edges are the rays through
		//
		//			(±w/c, ±h/c, 1, 1/n)
		//		and
		//			(±w/c, ±h/c, 1, 1/f)
		//
		//	to the clipping wedge whose edges are the rays through
		//
		//			(±1, ±1, 0, 1)
		//		and
		//			(±1, ±1, 1, 1)
		//
		//	Note that we require only that each ray on the view wedge
		//	map to the desired ray on the clipping wedge -- the given points
		//	do not have to match exactly.  For example, it's OK to map
		//	(w/h, h/w, 1, 1/n) to any positive multiple of (1, 1, 0, 1),
		//	it doesn't have to map to (1, 1, 0, 1) exactly.
		//
		//	To construct the required matrix, let's "follow our nose"
		//	and construct it as the product of several factors.
		//
		//		Note:  Unlike in older versions of the Geometry Games
		//		source code, matrices now act using
		//		the right-to-left (matrix)(column vector) convention,
		//		not the left-to-right (row vector)(matrix) convention.]
		//

		//	Factor #1
		//
		//		The quarter turn matrix
		//
		//			1  0  0  0
		//			0  1  0  0
		//			0  0  0 -1
		//			0  0  1  0
		//
		//		rotates
		//			(±w/c, ±h/c, 1, 1/n) and
		//			(±w/c, ±h/c, 1, 1/f)
		//		to
		//			(±w/c, ±h/c, -1/n, 1) and
		//			(±w/c, ±h/c, -1/f, 1)
		//
		
		//	Factor #2
		//
		//		The xy dilation
		//
		//			c/w  0   0   0
		//			 0  c/h  0   0
		//			 0   0   1   0
		//			 0   0   0   1
		//
		//		stretches or shrinks
		//			(±w/c, ±h/c, -1/n, 1) and
		//			(±w/c, ±h/c, -1/f, 1)
		//		to
		//			(±1, ±1, -1/n, 1) and
		//			(±1, ±1, -1/f, 1)
		//
		
		//	Factor #3
		//
		//		The z dilation
		//
		//			1  0  0  0
		//			0  1  0  0
		//			0  0  a  0
		//			0  0  0  1
		//
		//		with
		//			a = n * f / (f - n)
		//		stretches or shrinks the box to have unit length in the z direction, taking
		//			(±1, ±1, -1/n, 1) and
		//			(±1, ±1, -1/f, 1)
		//		to
		//			(±1, ±1, -f/(f - n), 1) and
		//			(±1, ±1, -n/(f - n), 1)
		//
		let a = n	//	= n*f/(f - n)  as f → ∞

		//	Factor #4
		//
		//		The shear
		//
		//			1  0  0  0
		//			0  1  0  0
		//			0  0  1  b
		//			0  0  0  1
		//
		//		with
		//			b = f / (f - n),
		//		translates the hyperplane w = 1 just the right amount to take
		//			(±1, ±1, -f/(f - n), 1) and
		//			(±1, ±1, -n/(f - n), 1)
		//		to
		//			(±1, ±1, 0, 1) and
		//			(±1, ±1, 1, 1)
		//
		//		as required.
		//
		let b = 1.0	//	= f/(f-n)  as f → ∞

		//	The projection matrix is the product (taken left-to-right!)
		//	of those four factors:
		//
		//			( 1  0  0  0 )( 1  0  0  0 )( c/w  0   0   0 )( 1  0  0  0 )
		//			( 0  1  0  0 )( 0  1  0  0 )(  0  c/h  0   0 )( 0  1  0  0 )
		//			( 0  0  1  b )( 0  0  a  0 )(  0   0   1   0 )( 0  0  0 -1 )
		//			( 0  0  0  1 )( 0  0  0  1 )(  0   0   0   1 )( 0  0  1  0 )
		//		=
		//			( c/w  0   0   0 )
		//			(  0  c/h  0   0 )
		//			(  0   0   b  -a )
		//			(  0   0   1   0 )
		//

		let theProjectionMatrix = simd_double4x4(
			SIMD4<Double>(c/w, 0.0, 0.0, 0.0),
			SIMD4<Double>(0.0, c/h, 0.0, 0.0),
			SIMD4<Double>(0.0, 0.0,  b,  1.0),
			SIMD4<Double>(0.0, 0.0, -a,  0.0))
		
		return theProjectionMatrix
	}
	
	func primitiveTranslationsInCameraCoordinates(
		modelData: CrystalFlightModel
	) -> simd_double3x3 {
	
		//	The atom and bond meshes will already be transformed and lit
		//	to account for the user's position and orientation.
		//	All we need to do here is prepare the translations
		//	needed to generate the tiling.
		
		let theUserOrientationInverse = simd_double3x3(modelData.itsUserOrientation.inverse)
		let thePrimitiveTranslationsInCameraCoordinates
			= theUserOrientationInverse * modelData.itsCrystal.primitiveTranslations

		return thePrimitiveTranslationsInCameraCoordinates
	}
	
	func viewFrustumOutwardUnitNormals(
		frameWidth: Int,
		frameHeight: Int
	) -> simd_double4x3 {	//	four 3-component vectors
	
		//	Compute the outward-pointing unit normal vectors
		//	to the four faces of the view frustum.
		
		//	By construction, the characteristic view size is the distance
		//	in the view, measured from the center of the view outwards,
		//	that subtends a 45° angle in the observer's field-of-view.
		//	An equivalent -- but more useful -- way to think about this
		//	is to imagine the user sitting characteristic-view-size units
		//	away from the center of the display.
		let w = Double(max(frameWidth,  1))	//	frameWidth = 0 should never occur
		let h = Double(max(frameHeight, 1))	//	frameHeight = 0 should never occur
		let c = characteristicViewSize(frameWidth: w, frameHeight: h)
		
		let n₀ = SIMD3<Double>(
					-c / sqrt(c*c + w*w),
					 0.0,
					-w / sqrt(c*c + w*w))
		let n₁ = SIMD3<Double>(
					+c / sqrt(c*c + w*w),
					 0.0,
					-w / sqrt(c*c + w*w))
		let n₂ = SIMD3<Double>(
					 0.0,
					-c / sqrt(c*c + h*h),
					-h / sqrt(c*c + h*h))
		let n₃ = SIMD3<Double>(
					 0.0,
					+c / sqrt(c*c + h*h),
					-h / sqrt(c*c + h*h))
		
		return simd_double4x3(n₀, n₁, n₂, n₃)
	}
	
	func latticeBounds(
		modelData: CrystalFlightModel,
		primitiveTranslationsInCameraCoordinates: simd_double3x3,
		boundingSphereInCameraCoordinates: Sphere
	) -> [ClosedRange<Int>] {

		//	If the center of a copy of the bounding sphere
		//	lies beyond theAdjustedHorizonRadius, then
		//	the whole bounding sphere lies beyond itsHorizonRadius.
		let theAdjustedHorizonRadius = modelData.itsHorizonRadius + boundingSphereInCameraCoordinates.radius

		//	How big a lattice should we iterate over to ensure
		//	that we don't miss any visible copies of the bounding sphere?
		
		//	Compute the dual basis.
		//
		//	Geometrically, theDualBasis[i] is orthogonal to the plane spanned by
		//
		//		theTranslations[ (i+1) mod 3 ]
		//		theTranslations[ (i+2) mod 3 ]
		//
		//	and
		//
		//		theDualBasis[i] · theTranslations[i] = 1
		//
		let theDualBasis = primitiveTranslationsInCameraCoordinates.inverse.transpose
				
		//	Dotting with theDualBasis[i] gives us a convenient way
		//	to measure distances from the orthogonal lattice plane,
		//	in units equal to the thickness of the "slabs"
		//	in the parallelepiped in that direction.  In particular,
		//	theDualBasis helps us decide the maximum number of "steps"
		//	we need to take along each of the three primitive directions.
		//
		let theLatticeBounds: [ClosedRange<Int>] = (0...2).map({ i in
		
			//	Note the bounding sphere center's "height" in the direction
			//	of theDualBasis[i], in units of the slab thickness.
			let theBoundingSphereHeightInSlabUnits
				= dot(boundingSphereInCameraCoordinates.center, theDualBasis[i])
			
			//	Convert theAdjustedHorizonRadius from ångströms
			//	to units of the slab thickness.
			let theAdjustedHorizonRadiusInSlabUnits
				= theAdjustedHorizonRadius * length(theDualBasis[i])
		
			//	How many slab units can we translate the bounding sphere
			//	before it -- and all its translates in the "horizontal" direction --
			//	lie beyond theAdjustedHorizonRadius?
			//
			let theBounds =
				(-Int(floor( theBoundingSphereHeightInSlabUnits - (-theAdjustedHorizonRadiusInSlabUnits))))
				...
				( Int(floor(theAdjustedHorizonRadiusInSlabUnits -   theBoundingSphereHeightInSlabUnits  )))
			
			return theBounds
		})
		
		return theLatticeBounds
	}

	
	func transformAndLightOneSet(
		computeEncoder: MTLComputeCommandEncoder,
		unlitSet: [GeometryGamesMeshBuffers<CrystalFlightUnlitVertex, CrystalFlightMeshIndexType>],
		litVertices: inout [MTLBuffer]	//	We may need to replace these buffers with larger ones
										//	when the user changes crystals.
	) {

		//	Note:  If CrystalFlight someday supports only A11 GPUs and later,
		//	I can delete theNonuniformThreadGroupSizesAreAvailable along
		//	with the fallback code that it invokes.
		//
		let theNonuniformThreadGroupSizesAreAvailable = itsDevice.supportsFamily(.apple4)	//	A11 GPU and later

		//	theLevelOfDetailColors get used only with gColorWithLevelOfDetail
		var theLevelOfDetailColors = [
			SIMD3<Float16>(0.0, 1.0, 1.0),	//	cyan
			SIMD3<Float16>(1.0, 0.0, 0.0),	//	red
			SIMD3<Float16>(1.0, 1.0, 0.0),	//	yellow
			SIMD3<Float16>(0.0, 1.0, 0.0) ]	//	green

		for theLevel in 0 ..< gNumLevelsOfDetail {

			let theUnlitMesh = unlitSet[theLevel]
			
			//	When the user changes crystals, we may need to replace
			//	the litVertices buffer with a bigger one.
			let theRequiredLitVertexBufferSize
				= theUnlitMesh.vertexCount * MemoryLayout<CrystalFlightLitVertex>.stride
			if litVertices[theLevel].length < theRequiredLitVertexBufferSize {

				guard let theNewBuffer = itsDevice.makeBuffer(
					length: theRequiredLitVertexBufferSize,
					options: .storageModePrivate)
				else { preconditionFailure("transformAndLightOneSet() failed to create theNewBuffer") }

				litVertices[theLevel] = theNewBuffer
			}
			let theTransformedAndLitVertexBuffer = litVertices[theLevel]
			
			computeEncoder.setBuffer(
				theUnlitMesh.vertexBuffer,
				offset: 0,
				index: BufferIndexUnlitVertices)
			
			computeEncoder.setBuffer(
				theTransformedAndLitVertexBuffer,
				offset: 0,
				index: BufferIndexLitVertices)

			//	During development, it's useful to see which
			//	images of the unit cell are being drawn
			//	at each level of detail.
			if itsColorByLevelOfDetail {
			
				computeEncoder.setBytes(
					&theLevelOfDetailColors[theLevel],
					length: MemoryLayout<SIMD3<Float16>>.stride,
					index: BufferIndexLevelOfDetailColor)
			}

			if theNonuniformThreadGroupSizesAreAvailable {

				//	Setting the threadgroup size to equal the threadExecutionWidth
				//	is a simple and effective choice.  In principle it could be
				//	some multiple of the threadExecutionWidth, but as expected
				//	the larger threadgroups tend to make the computation *slower*
				//	in this simple situation, because (I'm guessing -- don't know
				//	for sure) the larger threadgroups force the threads to share
				//	a smaller number of GPU cores, leaving other GPU cores idle.
				//	We could probably optimize this computation by using
				//	smaller threadgroups for the lower refinement levels
				//	(which involve relatively few vertices) and larger threadgroups
				//	for higher refinement levels (with lots of vertices),
				//	but such optimizations hardly seem worthwhile for a computation
				//	that's almost instantaneously fast in any case.
				//
				//	Here's the total time spent in the compute shader
				//	on my iPhone SE (2020), for various choices of the threadgroup size:
				//
				//		time	threadgroup size
				//		-----	--------------------------------
				//		56 µs	  32 =  1 × threadExecutionWidth
				//		51 µs	  64 =  2 × threadExecutionWidth
				//		62 µs	  96 =  3 × threadExecutionWidth
				//		78 µs	 128 =  4 × threadExecutionWidth
				//		...
				//		65 µs	1024 = 32 × threadExecutionWidth
				//
				computeEncoder.dispatchThreads(
					MTLSizeMake(theUnlitMesh.vertexCount, 1, 1),
					threadsPerThreadgroup: MTLSizeMake(itsTransformationAndLightingPipelineState.threadExecutionWidth, 1, 1))

			} else {	//	non-uniform threadgroup sizes not available

				//	On older GPUs that don't support non-uniform threadgroup sizes,
				//	use 1 thread per threadgroup.  Naively one might think that each
				//	such thread would occupy only 1 of the 32 possible SIMD slots
				//	in a GPU execution unit at run time, leaving the remaining 31 slots
				//	unused, but in practice this call is quite fast.  On the same
				//	iPhone SE (2020) used in the profiling results quoted above,
				//	this call takes
				//
				//		time	threadgroup size
				//		-----	--------------------------------
				//		64 µs	1 = (1/32) × threadExecutionWidth
				//
				//	Some (unconfirmed!) guesses about why the performance
				//	is so surprisingly good:  Maybe the GPU is scheduling
				//	32 different threadgroups to all run concurrently
				//	on the same GPU core?  Or maybe the whole process
				//	is I/O limited and the computational efficiency
				//	is irrelevant??
				//
				computeEncoder.dispatchThreadgroups(
					MTLSizeMake(theUnlitMesh.vertexCount, 1, 1),
					threadsPerThreadgroup: MTLSizeMake(1, 1, 1))
			}
		}
	}
}


// MARK: -
// MARK: Characteristic size

func characteristicViewSize(
	frameWidth: Double,	//	typically in pixels or points
	frameHeight: Double	//	typically in pixels or points
) -> Double {

	//	Here in Crystal Flight (but not necessarily in the other Geometry Games)
	//	this characteristicViewSize() function returns the distance
	//	in the view, measured from the center of the view outwards,
	//	that subtends a 45° angle in the observer's field-of-view.
	//
	//		····+----------+····
	//		    |         /
	//		    |       /
	//		    |     /
	//		    |45°/
	//		    | /
	//			*
	//		observer
	//
	//	The rendering code and the touch-handling code
	//	both rely on this same definition of the characteristic view size,
	//	to guarantee consistenty.

	//	I prefer assigning a ±45° field of view
	//	to the half-width or the half-height, whichever is bigger.
	//	This limits the field of view to 90°, which
	//	limits the perspective distortion caused by the fact
	//	that the user typically views the scene from a point
	//	much further back than the perspectively correct viewpoint.
	//	(If we rendered the scene from the user's true vantage point,
	//	the field of view would be far too narrow.)
	//
	return 0.5 * max(frameWidth, frameHeight)
}
